@@ -1,7 +1,7 @@
 -- We test cache so AM doesn't matter
 -- Just checking there are no mem debug WARNINGs during the cache population
 
-CREATE EXTENSION pg_tde;
+CREATE EXTENSION IF NOT EXISTS pg_tde;
 
 SELECT pg_tde_add_key_provider_file('file-vault','/tmp/pg_tde_test_keyring.per');
 SELECT pg_tde_set_principal_key('test-db-principal-key','file-vault');
