@@ -30,6 +30,24 @@
 #include "utils/typcache.h"
 #include "encryption/enc_tde.h"
 
+/* heap tuple residing in a buffer */
+typedef struct
+{
+	pg_node_attr(abstract)
+
+	HeapTupleTableSlot base;
+
+	/*
+	 * If buffer is not InvalidBuffer, then the slot is holding a pin on the
+	 * indicated buffer page; drop the pin when we release the slot's
+	 * reference to that buffer.  (TTS_FLAG_SHOULDFREE should not be set in
+	 * such a case, since presumably base.tuple is pointing into the buffer.)
+	 */
+	Buffer		buffer;			/* tuple's buffer, or InvalidBuffer */
+	char		decrypted_buffer[BLCKSZ];
+	InternalKey *cached_relation_key;
+} TDEBufferHeapTupleTableSlot;
+
 /*
  * TTSOpsTDEBufferHeapTuple is effectively the same as TTSOpsBufferHeapTuple slot.
  * The only difference is that it keeps the reference of the decrypted tuple
@@ -44,7 +62,7 @@ static inline void tdeheap_tts_buffer_heap_store_tuple(TupleTableSlot *slot,
 													   HeapTuple tuple,
 													   Buffer buffer,
 													   bool transfer_pin);
-static inline RelKeyData *get_current_slot_relation_key(TDEBufferHeapTupleTableSlot *bslot, Relation rel);
+static inline InternalKey *get_current_slot_relation_key(TDEBufferHeapTupleTableSlot *bslot, Relation rel);
 static void
 tdeheap_tts_buffer_heap_init(TupleTableSlot *slot)
 {
@@ -56,7 +74,7 @@ tdeheap_tts_buffer_heap_init(TupleTableSlot *slot)
 static void
 tdeheap_tts_buffer_heap_release(TupleTableSlot *slot)
 {
-	// nop
+	/* nop */
 }
 
 static void
@@ -242,12 +260,12 @@ tdeheap_tts_buffer_heap_copyslot(TupleTableSlot *dstslot, TupleTableSlot *srcslo
 		bdstslot->base.tuple = &bdstslot->base.tupdata;
 
 		/*
-		 * copy the decrypted buffer content as well
-		 * We only need to copy buffer upto tuple size
+		 * copy the decrypted buffer content as well We only need to copy
+		 * buffer upto tuple size
 		 */
 		memcpy(bdstslot->decrypted_buffer, bsrcslot->decrypted_buffer, HEAPTUPLESIZE + bsrcslot->base.tuple->t_len);
 		slot_copytuple(bdstslot->decrypted_buffer, bsrcslot->base.tuple);
-		bdstslot->base.tuple->t_data = ((HeapTuple)bdstslot->decrypted_buffer)->t_data;
+		bdstslot->base.tuple->t_data = ((HeapTuple) bdstslot->decrypted_buffer)->t_data;
 	}
 }
 
@@ -359,16 +377,16 @@ static pg_attribute_always_inline void
 tdeheap_slot_deform_heap_tuple(TupleTableSlot *slot, HeapTuple tuple, uint32 *offp,
 							   int natts)
 {
-	TupleDesc tupleDesc = slot->tts_tupleDescriptor;
-	Datum *values = slot->tts_values;
-	bool *isnull = slot->tts_isnull;
+	TupleDesc	tupleDesc = slot->tts_tupleDescriptor;
+	Datum	   *values = slot->tts_values;
+	bool	   *isnull = slot->tts_isnull;
 	HeapTupleHeader tup = tuple->t_data;
-	bool hasnulls = HeapTupleHasNulls(tuple);
-	int attnum;
-	char *tp;					/* ptr to tuple data */
-	uint32 off;					/* offset in tuple data */
-	bits8 *bp = tup->t_bits;	/* ptr to null bitmap in tuple */
-	bool slow;					/* can we use/set attcacheoff? */
+	bool		hasnulls = HeapTupleHasNulls(tuple);
+	int			attnum;
+	char	   *tp;				/* ptr to tuple data */
+	uint32		off;			/* offset in tuple data */
+	bits8	   *bp = tup->t_bits;	/* ptr to null bitmap in tuple */
+	bool		slow;			/* can we use/set attcacheoff? */
 
 	/* We can only fetch as many attributes as the tuple has. */
 	natts = Min(HeapTupleHeaderGetNatts(tuple->t_data), natts);
@@ -458,7 +476,7 @@ tdeheap_slot_deform_heap_tuple(TupleTableSlot *slot, HeapTuple tuple, uint32 *of
 static HeapTuple
 slot_copytuple(void *buffer, HeapTuple tuple)
 {
-	HeapTuple newTuple;
+	HeapTuple	newTuple;
 
 	if (!HeapTupleIsValid(tuple) || tuple->t_data == NULL)
 		return NULL;
@@ -535,7 +553,7 @@ PGTdeExecStoreBufferHeapTuple(Relation rel,
 
 	if (rel->rd_rel->relkind != RELKIND_TOASTVALUE)
 	{
-		RelKeyData *key = get_current_slot_relation_key(bslot, rel);
+		InternalKey *key = get_current_slot_relation_key(bslot, rel);
 
 		Assert(key != NULL);
 
@@ -577,7 +595,7 @@ PGTdeExecStorePinnedBufferHeapTuple(Relation rel,
 
 	if (rel->rd_rel->relkind != RELKIND_TOASTVALUE)
 	{
-		RelKeyData *key = get_current_slot_relation_key(bslot, rel);
+		InternalKey *key = get_current_slot_relation_key(bslot, rel);
 
 		slot_copytuple(bslot->decrypted_buffer, tuple);
 		PG_TDE_DECRYPT_TUPLE_EX(tuple, (HeapTuple) bslot->decrypted_buffer, key, "ExecStorePinnedBuffer");
@@ -592,7 +610,7 @@ PGTdeExecStorePinnedBufferHeapTuple(Relation rel,
 	return slot;
 }
 
-static inline RelKeyData *
+static inline InternalKey *
 get_current_slot_relation_key(TDEBufferHeapTupleTableSlot *bslot, Relation rel)
 {
 	Assert(bslot != NULL);
