@@ -18,6 +18,7 @@
 #include "catalog/tde_principal_key.h"
 #include "access/skey.h"
 #include "utils/lsyscache.h"
+#include "utils/memutils.h"
 #include "utils/snapmgr.h"
 #include "utils/fmgroids.h"
 #include "common/pg_tde_utils.h"
@@ -58,6 +59,10 @@ static GenericKeyring *load_keyring_provider_from_record(KeyringProvideRecord *p
 static inline void get_keyring_infofile_path(char *resPath, Oid dbOid);
 static bool fetch_next_key_provider(int fd, off_t *curr_pos, KeyringProvideRecord *provider);
 
+static uint32 write_key_provider_info(KeyringProvideRecord *provider,
+									  Oid database_id, off_t position,
+									  bool error_if_exists, bool write_xlog);
+
 #ifdef FRONTEND
 
 static SimplePtrList *scan_key_provider_file(ProviderScanType scanType, void *scanKey, Oid dbOid);
@@ -67,19 +72,34 @@ static void simple_list_free(SimplePtrList *list);
 
 static List *scan_key_provider_file(ProviderScanType scanType, void *scanKey, Oid dbOid);
 
-PG_FUNCTION_INFO_V1(pg_tde_add_key_provider_internal);
-Datum pg_tde_add_key_provider_internal(PG_FUNCTION_ARGS);
+PG_FUNCTION_INFO_V1(pg_tde_add_key_provider);
+Datum		pg_tde_add_key_provider(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_add_global_key_provider);
+Datum		pg_tde_add_global_key_provider(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_change_key_provider);
+Datum		pg_tde_change_key_provider(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_change_global_key_provider);
+Datum		pg_tde_change_global_key_provider(PG_FUNCTION_ARGS);
+
+static Datum pg_tde_list_all_key_providers_internal(const char *fname, bool global, PG_FUNCTION_ARGS);
 
 PG_FUNCTION_INFO_V1(pg_tde_list_all_key_providers);
-Datum pg_tde_list_all_key_providers(PG_FUNCTION_ARGS);
+Datum		pg_tde_list_all_key_providers(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_list_all_global_key_providers);
+Datum		pg_tde_list_all_global_key_providers(PG_FUNCTION_ARGS);
+
+static Datum pg_tde_change_key_provider_internal(PG_FUNCTION_ARGS, Oid dbOid);
+
+static Datum pg_tde_add_key_provider_internal(PG_FUNCTION_ARGS, Oid dbOid);
 
 #define PG_TDE_LIST_PROVIDERS_COLS 4
 
 static void key_provider_startup_cleanup(int tde_tbl_count, XLogExtensionInstall *ext_info, bool redo, void *arg);
 static const char *get_keyring_provider_typename(ProviderType p_type);
-static uint32 write_key_provider_info(KeyringProvideRecord *provider, 
-									Oid database_id, off_t position, 
-									bool error_if_exists, bool write_xlog);
 
 static Size initialize_shared_state(void *start_address);
 static Size required_shared_mem_size(void);
@@ -140,21 +160,6 @@ key_provider_startup_cleanup(int tde_tbl_count, XLogExtensionInstall *ext_info,
 	cleanup_key_provider_info(ext_info->database_id);
 }
 
-ProviderType
-get_keyring_provider_from_typename(char *provider_type)
-{
-	if (provider_type == NULL)
-		return UNKNOWN_KEY_PROVIDER;
-
-	if (strcmp(FILE_KEYRING_TYPE, provider_type) == 0)
-		return FILE_KEY_PROVIDER;
-	if (strcmp(VAULTV2_KEYRING_TYPE, provider_type) == 0)
-		return VAULT_V2_KEY_PROVIDER;
-	if (strcmp(KMIP_KEYRING_TYPE, provider_type) == 0)
-		return KMIP_KEY_PROVIDER;
-	return UNKNOWN_KEY_PROVIDER;
-}
-
 static const char *
 get_keyring_provider_typename(ProviderType p_type)
 {
@@ -178,41 +183,240 @@ GetAllKeyringProviders(Oid dbOid)
 	return scan_key_provider_file(PROVIDER_SCAN_ALL, NULL, dbOid);
 }
 
-GenericKeyring *
-GetKeyProviderByName(const char *provider_name, Oid dbOid)
+uint32
+redo_key_provider_info(KeyringProviderXLRecord *xlrec)
 {
+	return write_key_provider_info(&xlrec->provider, xlrec->database_id, xlrec->offset_in_file, false, false);
+}
+
+void
+cleanup_key_provider_info(Oid databaseId)
+{
+	/* Remove the key provider info file */
+	char		kp_info_path[MAXPGPATH] = {0};
+
+	get_keyring_infofile_path(kp_info_path, databaseId);
+	PathNameDeleteTemporaryFile(kp_info_path, false);
+}
+
+Datum
+pg_tde_change_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_change_key_provider_internal(fcinfo, MyDatabaseId);
+}
+
+Datum
+pg_tde_change_global_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_change_key_provider_internal(fcinfo, GLOBAL_DATA_TDE_OID);
+}
+
+static Datum
+pg_tde_change_key_provider_internal(PG_FUNCTION_ARGS, Oid dbOid)
+{
+	char	   *provider_type = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	char	   *options = text_to_cstring(PG_GETARG_TEXT_PP(2));
+	int			nlen,
+				olen;
+	KeyringProvideRecord provider;
+
+	/* reports error if not found */
+	GenericKeyring *keyring = GetKeyProviderByName(provider_name, dbOid);
+
+	pfree(keyring);
+
+	nlen = strlen(provider_name);
+	if (nlen >= sizeof(provider.provider_name))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("too long provider name, maximum lenght is %ld bytes", sizeof(provider.provider_name) - 1)));
+
+	olen = strlen(options);
+	if (olen >= sizeof(provider.options))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("too large provider options, maximum size is %ld bytes", sizeof(provider.options) - 1)));
+
+	/* Struct will be saved to disk so keep clean */
+	memset(&provider, 0, sizeof(provider));
+	provider.provider_id = 0;
+	memcpy(provider.provider_name, provider_name, nlen);
+	memcpy(provider.options, options, olen);
+	provider.provider_type = get_keyring_provider_from_typename(provider_type);
+	modify_key_provider_info(&provider, dbOid, true);
+
+	PG_RETURN_INT32(provider.provider_id);
+}
+
+Datum
+pg_tde_add_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_add_key_provider_internal(fcinfo, MyDatabaseId);
+}
+
+Datum
+pg_tde_add_global_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_add_key_provider_internal(fcinfo, GLOBAL_DATA_TDE_OID);
+}
+
+Datum
+pg_tde_add_key_provider_internal(PG_FUNCTION_ARGS, Oid dbOid)
+{
+	char	   *provider_type = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	char	   *options = text_to_cstring(PG_GETARG_TEXT_PP(2));
+	int			nlen,
+				olen;
+	KeyringProvideRecord provider;
+
+	nlen = strlen(provider_name);
+	if (nlen >= sizeof(provider.provider_name) - 1)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("too long provider name, maximum lenght is %ld bytes", sizeof(provider.provider_name) - 1)));
+
+	olen = strlen(options);
+	if (olen >= sizeof(provider.options))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("too large provider options, maximum size is %ld bytes", sizeof(provider.options) - 1)));
+
+	/* Struct will be saved to disk so keep clean */
+	memset(&provider, 0, sizeof(provider));
+	provider.provider_id = 0;
+	memcpy(provider.provider_name, provider_name, nlen);
+	memcpy(provider.options, options, olen);
+	provider.provider_type = get_keyring_provider_from_typename(provider_type);
+	save_new_key_provider_info(&provider, dbOid, true);
+
+	PG_RETURN_INT32(provider.provider_id);
+}
+
+Datum
+pg_tde_list_all_key_providers(PG_FUNCTION_ARGS)
+{
+	return pg_tde_list_all_key_providers_internal("pg_tde_list_all_key_providers", false, fcinfo);
+}
+
+Datum
+pg_tde_list_all_global_key_providers(PG_FUNCTION_ARGS)
+{
+	return pg_tde_list_all_key_providers_internal("pg_tde_list_all_key_providers_global", true, fcinfo);
+}
+
+static Datum
+pg_tde_list_all_key_providers_internal(const char *fname, bool global, PG_FUNCTION_ARGS)
+{
+	Oid			database = (global ? GLOBAL_DATA_TDE_OID : MyDatabaseId);
+	List	   *all_providers = GetAllKeyringProviders(database);
+	ListCell   *lc;
+	Tuplestorestate *tupstore;
+	TupleDesc	tupdesc;
+	MemoryContext per_query_ctx;
+	MemoryContext oldcontext;
+	ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
+
+	/* check to see if caller supports us returning a tuplestore */
+	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("%s: set-valued function called in context that cannot accept a set", fname)));
+	if (!(rsinfo->allowedModes & SFRM_Materialize))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("%s: materialize mode required, but it is not allowed in this context", fname)));
+
+	/* Switch into long-lived context to construct returned data structures */
+	per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
+	oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
+	/* Build a tuple descriptor for our result type */
+	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+		elog(ERROR, "%s: return type must be a row type", fname);
+
+	tupstore = tuplestore_begin_heap(true, false, work_mem);
+	rsinfo->returnMode = SFRM_Materialize;
+	rsinfo->setResult = tupstore;
+	rsinfo->setDesc = tupdesc;
+
+	MemoryContextSwitchTo(oldcontext);
+
+	foreach(lc, all_providers)
+	{
+		Datum		values[PG_TDE_LIST_PROVIDERS_COLS] = {0};
+		bool		nulls[PG_TDE_LIST_PROVIDERS_COLS] = {0};
+		GenericKeyring *keyring = (GenericKeyring *) lfirst(lc);
+		int			i = 0;
+
+		values[i++] = Int32GetDatum(keyring->keyring_id);
+		values[i++] = CStringGetTextDatum(keyring->provider_name);
+		values[i++] = CStringGetTextDatum(get_keyring_provider_typename(keyring->type));
+		values[i++] = CStringGetTextDatum(keyring->options);
+		tuplestore_putvalues(tupstore, tupdesc, values, nulls);
+
+		debug_print_kerying(keyring);
+	}
+	list_free_deep(all_providers);
+	return (Datum) 0;
+}
+
+GenericKeyring *
+GetKeyProviderByID(int provider_id, Oid dbOid)
+{
+	Oid			realOid = provider_id < 0 ? GLOBAL_DATA_TDE_OID : dbOid;
 	GenericKeyring *keyring = NULL;
-	List *providers = scan_key_provider_file(PROVIDER_SCAN_BY_NAME, (void *) provider_name, dbOid);
+	List	   *providers = scan_key_provider_file(PROVIDER_SCAN_BY_ID, &provider_id, realOid);
 
 	if (providers != NIL)
 	{
 		keyring = (GenericKeyring *) linitial(providers);
 		list_free(providers);
 	}
-	else
-	{
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("key provider \"%s\" does not exists", provider_name),
-				 errhint("Use pg_tde_add_key_provider interface to create the key provider")));
-	}
 	return keyring;
 }
 
+#endif							/* !FRONTEND */
 
 static uint32
 write_key_provider_info(KeyringProvideRecord *provider, Oid database_id,
 						off_t position, bool error_if_exists, bool write_xlog)
 {
-	off_t bytes_written = 0;
-	off_t curr_pos = 0;
-	int fd;
-	int max_provider_id = 0;
-	char kp_info_path[MAXPGPATH] = {0};
+	off_t		bytes_written = 0;
+	off_t		curr_pos = 0;
+	int			fd;
+	int			seek_pos = -1;
+
+	/* Named max, but global key provider oids are stored as negative numbers! */
+	int			max_provider_id = 0;
+	char		kp_info_path[MAXPGPATH] = {0};
 	KeyringProvideRecord existing_provider;
+	GenericKeyring *record;
 
 	Assert(provider != NULL);
 
+	if (error_if_exists && provider->provider_id != 0)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_DATA_EXCEPTION), errmsg("Invalid write provider call")));
+	}
+
+	/* Try to parse the JSON data first: if it doesn't work, don't save it! */
+	if (provider->provider_type != UNKNOWN_KEY_PROVIDER)
+	{
+		record = load_keyring_provider_from_record(provider);
+		if (record == NULL)
+		{
+			ereport(ERROR,
+					(errcode(ERRCODE_DATA_EXCEPTION), errmsg("Invalid provider options")));
+		}
+		else
+		{
+			pfree(record);
+		}
+	}
+
 	get_keyring_infofile_path(kp_info_path, database_id);
 
 	LWLockAcquire(tde_provider_info_lock(), LW_EXCLUSIVE);
@@ -231,27 +435,53 @@ write_key_provider_info(KeyringProvideRecord *provider, Oid database_id,
 		 * we also need to verify the name conflict and generate the next
 		 * provider ID
 		 */
+		int			before_pos = curr_pos;
+
 		while (fetch_next_key_provider(fd, &curr_pos, &existing_provider))
 		{
-			if (strcmp(existing_provider.provider_name, provider->provider_name) == 0)
+			if (provider->provider_id != 0 && existing_provider.provider_id == provider->provider_id)
 			{
-				close(fd);
-				LWLockRelease(tde_provider_info_lock());
-				ereport(error_if_exists ? ERROR : DEBUG1,
-						(errcode(ERRCODE_DUPLICATE_OBJECT),
-						errmsg("key provider \"%s\" already exists", provider->provider_name)));
-
-				if (!error_if_exists)
+				seek_pos = before_pos;
+				break;
+			}
+			if (strlen(existing_provider.provider_name) > 0 && strcmp(existing_provider.provider_name, provider->provider_name) == 0)
+			{
+				if (error_if_exists)
 				{
+					close(fd);
+					LWLockRelease(tde_provider_info_lock());
+					ereport(ERROR,
+							(errcode(ERRCODE_DUPLICATE_OBJECT),
+							 errmsg("key provider \"%s\" already exists", provider->provider_name)));
+				}
+				else
+				{
+
+					seek_pos = before_pos;
 					provider->provider_id = existing_provider.provider_id;
-					return provider->provider_id;
+					break;
 				}
 			}
-			if (max_provider_id < existing_provider.provider_id)
-				max_provider_id = existing_provider.provider_id;
+			if (max_provider_id < abs(existing_provider.provider_id))
+				max_provider_id = abs(existing_provider.provider_id);
+
+			before_pos = curr_pos;
 		}
-		provider->provider_id = max_provider_id + 1;
-		curr_pos = lseek(fd, 0, SEEK_END);
+		if (seek_pos == -1)
+		{
+			provider->provider_id = max_provider_id + 1;
+
+			if (database_id == GLOBAL_DATA_TDE_OID)
+			{
+				provider->provider_id = -provider->provider_id;
+			}
+			curr_pos = lseek(fd, 0, SEEK_END);
+		}
+		else
+		{
+			curr_pos = lseek(fd, seek_pos, SEEK_CUR);
+		}
+
 
 		/*
 		 * emit the xlog here. So that we can handle partial file write errors
@@ -259,6 +489,7 @@ write_key_provider_info(KeyringProvideRecord *provider, Oid database_id,
 		 */
 		if (write_xlog)
 		{
+#ifndef FRONTEND
 			KeyringProviderXLRecord xlrec;
 
 			xlrec.database_id = database_id;
@@ -268,6 +499,9 @@ write_key_provider_info(KeyringProvideRecord *provider, Oid database_id,
 			XLogBeginInsert();
 			XLogRegisterData((char *) &xlrec, sizeof(KeyringProviderXLRecord));
 			XLogInsert(RM_TDERMGR_ID, XLOG_TDE_ADD_KEY_PROVIDER_KEY);
+#else
+			Assert(0);
+#endif
 		}
 	}
 	else
@@ -307,126 +541,40 @@ write_key_provider_info(KeyringProvideRecord *provider, Oid database_id,
 	return provider->provider_id;
 }
 
+
 /*
  * Save the key provider info to the file
  */
 uint32
-save_new_key_provider_info(KeyringProvideRecord* provider, Oid databaseId, bool write_xlog)
+save_new_key_provider_info(KeyringProvideRecord *provider, Oid databaseId, bool write_xlog)
 {
 	return write_key_provider_info(provider, databaseId, -1, true, write_xlog);
 }
 
 uint32
-redo_key_provider_info(KeyringProviderXLRecord *xlrec)
+modify_key_provider_info(KeyringProvideRecord *provider, Oid databaseId, bool write_xlog)
 {
-	return write_key_provider_info(&xlrec->provider, xlrec->database_id, xlrec->offset_in_file, true, false);
+	return write_key_provider_info(provider, databaseId, -1, false, write_xlog);
 }
 
-void
-cleanup_key_provider_info(Oid databaseId)
+uint32
+delete_key_provider_info(int provider_id, Oid databaseId, bool write_xlog)
 {
-	/* Remove the key provider info file */
-	char		kp_info_path[MAXPGPATH] = {0};
+	KeyringProvideRecord kpr;
 
-	get_keyring_infofile_path(kp_info_path, databaseId);
-	PathNameDeleteTemporaryFile(kp_info_path, false);
+	memset(&kpr, 0, sizeof(KeyringProvideRecord));
+	kpr.provider_id = provider_id;
+
+	return modify_key_provider_info(&kpr, databaseId, write_xlog);
 }
 
-Datum
-pg_tde_add_key_provider_internal(PG_FUNCTION_ARGS)
-{
-	char *provider_type = text_to_cstring(PG_GETARG_TEXT_PP(0));
-	char *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
-	char *options = text_to_cstring(PG_GETARG_TEXT_PP(2));
-	bool is_global = PG_GETARG_BOOL(3);
-	KeyringProvideRecord provider;
-	Oid dbOid = is_global ? GLOBAL_DATA_TDE_OID : MyDatabaseId;
-
-	strncpy(provider.options, options, sizeof(provider.options));
-	strncpy(provider.provider_name, provider_name, sizeof(provider.provider_name));
-	provider.provider_type = get_keyring_provider_from_typename(provider_type);
-	save_new_key_provider_info(&provider, dbOid, true);
-
-	PG_RETURN_INT32(provider.provider_id);
-}
-
-Datum
-pg_tde_list_all_key_providers(PG_FUNCTION_ARGS)
-{
-	List *all_providers = GetAllKeyringProviders(MyDatabaseId);
-	ListCell *lc;
-	Tuplestorestate *tupstore;
-	TupleDesc tupdesc;
-	MemoryContext per_query_ctx;
-	MemoryContext oldcontext;
-	ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
-
-	/* check to see if caller supports us returning a tuplestore */
-	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
-		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("pg_tde_list_all_key_providers: set-valued function called in context that cannot accept a set")));
-	if (!(rsinfo->allowedModes & SFRM_Materialize))
-		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("pg_tde_list_all_key_providers: materialize mode required, but it is not allowed in this context")));
-
-	/* Switch into long-lived context to construct returned data structures */
-	per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
-	oldcontext = MemoryContextSwitchTo(per_query_ctx);
-
-	/* Build a tuple descriptor for our result type */
-	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
-		elog(ERROR, "pg_tde_list_all_key_providers: return type must be a row type");
-
-	tupstore = tuplestore_begin_heap(true, false, work_mem);
-	rsinfo->returnMode = SFRM_Materialize;
-	rsinfo->setResult = tupstore;
-	rsinfo->setDesc = tupdesc;
-
-	MemoryContextSwitchTo(oldcontext);
-
-	foreach(lc, all_providers)
-	{
-		Datum values[PG_TDE_LIST_PROVIDERS_COLS] = {0};
-		bool nulls[PG_TDE_LIST_PROVIDERS_COLS] = {0};
-		GenericKeyring *keyring = (GenericKeyring *) lfirst(lc);
-		int i = 0;
-
-		values[i++] = Int32GetDatum(keyring->key_id);
-		values[i++] = CStringGetTextDatum(keyring->provider_name);
-		values[i++] = CStringGetTextDatum(get_keyring_provider_typename(keyring->type));
-		values[i++] = CStringGetTextDatum(keyring->options);
-		tuplestore_putvalues(tupstore, tupdesc, values, nulls);
-
-		debug_print_kerying(keyring);
-	}
-	list_free_deep(all_providers);
-	return (Datum) 0;
-}
-
-GenericKeyring *
-GetKeyProviderByID(int provider_id, Oid dbOid)
-{
-	GenericKeyring *keyring = NULL;
-	List *providers = scan_key_provider_file(PROVIDER_SCAN_BY_ID, &provider_id, dbOid);
-
-	if (providers != NIL)
-	{
-		keyring = (GenericKeyring *) linitial(providers);
-		list_free(providers);
-	}
-	return keyring;
-}
-
-#endif /* !FRONTEND */
-
 #ifdef FRONTEND
 GenericKeyring *
 GetKeyProviderByID(int provider_id, Oid dbOid)
 {
+	Oid			realOid = provider_id < 0 ? GLOBAL_DATA_TDE_OID : dbOid;
 	GenericKeyring *keyring = NULL;
-	SimplePtrList *providers = scan_key_provider_file(PROVIDER_SCAN_BY_ID, &provider_id, dbOid);
+	SimplePtrList *providers = scan_key_provider_file(PROVIDER_SCAN_BY_ID, &provider_id, realOid);
 
 	if (providers != NULL)
 	{
@@ -451,7 +599,7 @@ simple_list_free(SimplePtrList *list)
 		cell = next;
 	}
 }
-#endif /* FRONTEND */
+#endif							/* FRONTEND */
 
 /*
  * Scan the key provider info file and can also apply filter based on scanType
@@ -463,12 +611,12 @@ static SimplePtrList *
 #endif
 scan_key_provider_file(ProviderScanType scanType, void *scanKey, Oid dbOid)
 {
-	off_t curr_pos = 0;
-	int fd;
-	char kp_info_path[MAXPGPATH] = {0};
+	off_t		curr_pos = 0;
+	int			fd;
+	char		kp_info_path[MAXPGPATH] = {0};
 	KeyringProvideRecord provider;
 #ifndef FRONTEND
-	List *providers_list = NIL;
+	List	   *providers_list = NIL;
 #else
 	SimplePtrList *providers_list = NULL;
 #endif
@@ -493,6 +641,12 @@ scan_key_provider_file(ProviderScanType scanType, void *scanKey, Oid dbOid)
 	{
 		bool		match = false;
 
+		if (provider.provider_type == UNKNOWN_KEY_PROVIDER)
+		{
+			/* deleted provider */
+			continue;
+		}
+
 		ereport(DEBUG2,
 				(errmsg("read key provider ID=%d %s", provider.provider_id, provider.provider_name)));
 
@@ -543,10 +697,10 @@ load_keyring_provider_from_record(KeyringProvideRecord *provider)
 	keyring = load_keyring_provider_options(provider->provider_type, provider->options);
 	if (keyring)
 	{
-		keyring->key_id = provider->provider_id;
-		strncpy(keyring->provider_name, provider->provider_name, sizeof(keyring->provider_name));
+		keyring->keyring_id = provider->provider_id;
+		memcpy(keyring->provider_name, provider->provider_name, sizeof(keyring->provider_name));
 		keyring->type = provider->provider_type;
-		strncpy(keyring->options, provider->options, sizeof(keyring->options));
+		memcpy(keyring->options, provider->options, sizeof(keyring->options));
 		debug_print_kerying(keyring);
 	}
 	return keyring;
@@ -565,7 +719,7 @@ load_keyring_provider_options(ProviderType provider_type, char *keyring_options)
 			return (GenericKeyring *) load_vaultV2_keyring_provider_options(keyring_options);
 			break;
 		case KMIP_KEY_PROVIDER:
-			return (GenericKeyring *)load_kmip_keyring_provider_options(keyring_options);
+			return (GenericKeyring *) load_kmip_keyring_provider_options(keyring_options);
 			break;
 		default:
 			break;
@@ -586,7 +740,7 @@ load_file_keyring_provider_options(char *keyring_options)
 		return NULL;
 	}
 
-	if (strlen(file_keyring->file_name) == 0)
+	if (file_keyring->file_name == NULL || file_keyring->file_name[0] == '\0')
 	{
 		ereport(WARNING,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
@@ -610,16 +764,16 @@ load_vaultV2_keyring_provider_options(char *keyring_options)
 		return NULL;
 	}
 
-	if (strlen(vaultV2_keyring->vault_token) == 0 ||
-		strlen(vaultV2_keyring->vault_url) == 0 ||
-		strlen(vaultV2_keyring->vault_mount_path) == 0)
+	if (vaultV2_keyring->vault_token == NULL || vaultV2_keyring->vault_token[0] == '\0' ||
+		vaultV2_keyring->vault_url == NULL || vaultV2_keyring->vault_url[0] == '\0' ||
+		vaultV2_keyring->vault_mount_path == NULL || vaultV2_keyring->vault_mount_path[0] == '\0')
 	{
 		ereport(WARNING,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
 				 errmsg("missing in the keyring options:%s%s%s",
-						*(vaultV2_keyring->vault_token) ? "" : " token",
-						*(vaultV2_keyring->vault_url) ? "" : " url",
-						*(vaultV2_keyring->vault_mount_path) ? "" : " mountPath")));
+						(vaultV2_keyring->vault_token != NULL && vaultV2_keyring->vault_token[0] != '\0') ? "" : " token",
+						(vaultV2_keyring->vault_url != NULL && vaultV2_keyring->vault_url[0] != '\0') ? "" : " url",
+						(vaultV2_keyring->vault_mount_path != NULL && vaultV2_keyring->vault_mount_path[0] != '\0') ? "" : " mountPath")));
 		return NULL;
 	}
 
@@ -640,15 +794,17 @@ load_kmip_keyring_provider_options(char *keyring_options)
 	}
 
 	if (strlen(kmip_keyring->kmip_host) == 0 ||
+		strlen(kmip_keyring->kmip_port) == 0 ||
 		strlen(kmip_keyring->kmip_ca_path) == 0 ||
 		strlen(kmip_keyring->kmip_cert_path) == 0)
 	{
 		ereport(WARNING,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("missing in the keyring options:%s%s%s",
-						*(kmip_keyring->kmip_host) ? "" : " kmip_host",
-						*(kmip_keyring->kmip_ca_path) ? "" : " kmip_ca_path",
-						*(kmip_keyring->kmip_cert_path) ? "" : " kmip_cert_path")));
+				 errmsg("missing in the keyring options:%s%s%s%s",
+						(kmip_keyring->kmip_host != NULL && kmip_keyring->kmip_host[0] != '\0') ? "" : " host",
+						(kmip_keyring->kmip_port != NULL && kmip_keyring->kmip_port[0] != '\0') ? "" : " port",
+						(kmip_keyring->kmip_ca_path != NULL && kmip_keyring->kmip_ca_path[0] != '\0') ? "" : " caPath",
+						(kmip_keyring->kmip_cert_path != NULL && kmip_keyring->kmip_cert_path[0] != '\0') ? "" : " certPath")));
 		return NULL;
 	}
 
@@ -658,11 +814,11 @@ load_kmip_keyring_provider_options(char *keyring_options)
 static void
 debug_print_kerying(GenericKeyring *keyring)
 {
-	int debug_level = DEBUG2;
+	int			debug_level = DEBUG2;
 
 	elog(debug_level, "Keyring type: %d", keyring->type);
 	elog(debug_level, "Keyring name: %s", keyring->provider_name);
-	elog(debug_level, "Keyring id: %d", keyring->key_id);
+	elog(debug_level, "Keyring id: %d", keyring->keyring_id);
 	switch (keyring->type)
 	{
 		case FILE_KEY_PROVIDER:
@@ -674,12 +830,12 @@ debug_print_kerying(GenericKeyring *keyring)
 			elog(debug_level, "Vault Keyring Mount Path: %s", ((VaultV2Keyring *) keyring)->vault_mount_path);
 			elog(debug_level, "Vault Keyring CA Path: %s", ((VaultV2Keyring *) keyring)->vault_ca_path);
 			break;
-       case KMIP_KEY_PROVIDER:                                                                                                   
-            elog(debug_level, "KMIP Keyring Host: %s", ((KmipKeyring *)keyring)->kmip_host);                                  
-            elog(debug_level, "KMIP Keyring Port: %s", ((KmipKeyring *)keyring)->kmip_port);                                  
-            elog(debug_level, "KMIP Keyring CA Path: %s", ((KmipKeyring *)keyring)->kmip_ca_path);                            
-            elog(debug_level, "KMIP Keyring Cert Path: %s", ((KmipKeyring *)keyring)->kmip_cert_path);                        
-            break; 
+		case KMIP_KEY_PROVIDER:
+			elog(debug_level, "KMIP Keyring Host: %s", ((KmipKeyring *) keyring)->kmip_host);
+			elog(debug_level, "KMIP Keyring Port: %s", ((KmipKeyring *) keyring)->kmip_port);
+			elog(debug_level, "KMIP Keyring CA Path: %s", ((KmipKeyring *) keyring)->kmip_ca_path);
+			elog(debug_level, "KMIP Keyring Cert Path: %s", ((KmipKeyring *) keyring)->kmip_cert_path);
+			break;
 		case UNKNOWN_KEY_PROVIDER:
 			elog(debug_level, "Unknown Keyring ");
 			break;
@@ -698,7 +854,7 @@ get_keyring_infofile_path(char *resPath, Oid dbOid)
 static bool
 fetch_next_key_provider(int fd, off_t *curr_pos, KeyringProvideRecord *provider)
 {
-	off_t bytes_read = 0;
+	off_t		bytes_read = 0;
 
 	Assert(provider != NULL);
 	Assert(fd >= 0);
@@ -719,3 +875,51 @@ fetch_next_key_provider(int fd, off_t *curr_pos, KeyringProvideRecord *provider)
 	}
 	return true;
 }
+
+ProviderType
+get_keyring_provider_from_typename(char *provider_type)
+{
+	if (provider_type == NULL)
+		return UNKNOWN_KEY_PROVIDER;
+
+	if (strcmp(FILE_KEYRING_TYPE, provider_type) == 0)
+		return FILE_KEY_PROVIDER;
+	if (strcmp(VAULTV2_KEYRING_TYPE, provider_type) == 0)
+		return VAULT_V2_KEY_PROVIDER;
+	if (strcmp(KMIP_KEYRING_TYPE, provider_type) == 0)
+		return KMIP_KEY_PROVIDER;
+	return UNKNOWN_KEY_PROVIDER;
+}
+
+GenericKeyring *
+GetKeyProviderByName(const char *provider_name, Oid dbOid)
+{
+	GenericKeyring *keyring = NULL;
+
+#ifndef FRONTEND
+	static List *providers;
+#else
+	static SimplePtrList *providers;
+#endif
+
+	providers = scan_key_provider_file(PROVIDER_SCAN_BY_NAME, (void *) provider_name, dbOid);
+
+	if (providers != NULL)
+	{
+#ifndef FRONTEND
+		keyring = (GenericKeyring *) linitial(providers);
+		list_free(providers);
+#else
+		keyring = (GenericKeyring *) providers->head->ptr;
+		simple_list_free(providers);
+#endif
+	}
+	else
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("key provider \"%s\" does not exists", provider_name),
+				 errhint("Use pg_tde_add_key_provider interface to create the key provider")));
+	}
+	return keyring;
+}
