@@ -25,27 +25,39 @@
 #include "access/pg_tde_xlog.h"
 #include <sys/mman.h>
 #include <sys/time.h>
+#include "utils/fmgroids.h"
+#include "utils/guc.h"
+#include "catalog/pg_database.h"
 
 #include "access/pg_tde_tdemap.h"
 #include "catalog/tde_global_space.h"
 #ifndef FRONTEND
+#include "access/genam.h"
+#include "access/table.h"
 #include "common/pg_tde_shmem.h"
 #include "funcapi.h"
 #include "storage/lwlock.h"
 #else
 #include "pg_tde_fe.h"
 #endif
+#include "pg_tde_guc.h"
 
 #include <sys/time.h>
 
 #ifndef FRONTEND
 
+PG_FUNCTION_INFO_V1(pg_tde_delete_key_provider);
+PG_FUNCTION_INFO_V1(pg_tde_delete_global_key_provider);
+
+PG_FUNCTION_INFO_V1(pg_tde_verify_principal_key);
+PG_FUNCTION_INFO_V1(pg_tde_verify_global_principal_key);
+
 typedef struct TdePrincipalKeySharedState
 {
 	LWLockPadded *Locks;
-	int	hashTrancheId;
+	int			hashTrancheId;
 	dshash_table_handle hashHandle;
-	void *rawDsaArea;		/* DSA area pointer */
+	void	   *rawDsaArea;		/* DSA area pointer */
 
 } TdePrincipalKeySharedState;
 
@@ -75,18 +87,45 @@ static void shared_memory_shutdown(int code, Datum arg);
 static void principal_key_startup_cleanup(int tde_tbl_count, XLogExtensionInstall *ext_info, bool redo, void *arg);
 static void clear_principal_key_cache(Oid databaseId);
 static inline dshash_table *get_principal_key_Hash(void);
-static TDEPrincipalKey *get_principal_key_from_keyring(Oid dbOid);
 static TDEPrincipalKey *get_principal_key_from_cache(Oid dbOid);
+static bool pg_tde_is_same_principal_key(TDEPrincipalKey *a, TDEPrincipalKey *b);
+static void pg_tde_update_global_principal_key_everywhere(TDEPrincipalKey *oldKey, TDEPrincipalKey *newKey);
 static void push_principal_key_to_cache(TDEPrincipalKey *principalKey);
 static Datum pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid);
-static keyInfo *load_latest_versioned_key_name(TDEPrincipalKeyInfo *principal_key_info,
-											   GenericKeyring *keyring,
-											   bool ensure_new_key);
-static TDEPrincipalKey *set_principal_key_with_keyring(const char *key_name,
-													   GenericKeyring *keyring,
-													   Oid dbOid,
-													   bool ensure_new_key);
-static TDEPrincipalKey *alter_keyprovider_for_principal_key(GenericKeyring *newKeyring,Oid dbOid);
+static bool set_principal_key_with_keyring(const char *key_name,
+										   const char *provider_name,
+										   Oid providerOid,
+										   Oid dbOid,
+										   bool ensure_new_key);
+static bool pg_tde_is_provider_used(Oid databaseOid, Oid providerId);
+static bool pg_tde_verify_principal_key_internal(Oid databaseOid);
+
+static Datum pg_tde_delete_key_provider_internal(PG_FUNCTION_ARGS, int is_global);
+
+PG_FUNCTION_INFO_V1(pg_tde_set_default_principal_key);
+Datum		pg_tde_set_default_principal_key(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_set_principal_key);
+Datum		pg_tde_set_principal_key(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_set_global_principal_key);
+Datum		pg_tde_set_principal_key(PG_FUNCTION_ARGS);
+
+PG_FUNCTION_INFO_V1(pg_tde_set_server_principal_key);
+Datum		pg_tde_set_principal_key(PG_FUNCTION_ARGS);
+
+enum global_status
+{
+	GS_LOCAL,
+	GS_GLOBAL,
+	GS_SERVER,
+	GS_DEFAULT
+};
+
+static Datum
+			pg_tde_set_principal_key_internal(char *principal_key_name, enum global_status global, char *provider_name, bool ensure_new_key);
+
+
 
 static const TDEShmemSetupRoutine principal_key_info_shmem_routine = {
 	.init_shared_state = initialize_shared_state,
@@ -210,7 +249,7 @@ shared_memory_shutdown(int code, Datum arg)
 }
 
 bool
-save_principal_key_info(TDEPrincipalKeyInfo *principal_key_info)
+create_principal_key_info(TDEPrincipalKeyInfo *principal_key_info)
 {
 	Assert(principal_key_info != NULL);
 
@@ -224,228 +263,136 @@ update_principal_key_info(TDEPrincipalKeyInfo *principal_key_info)
 	return pg_tde_save_principal_key(principal_key_info, false, true);
 }
 
-/*
- * SetPrincipalkey:
- * We need to ensure that only one principal key is set for a database.
- */
-TDEPrincipalKey *
-set_principal_key_with_keyring(const char *key_name, GenericKeyring *keyring,
-							   Oid dbOid, bool ensure_new_key)
+bool
+set_principal_key_with_keyring(const char *key_name, const char *provider_name,
+							   Oid providerOid, Oid dbOid, bool ensure_new_key)
 {
-	TDEPrincipalKey *principalKey = NULL;
+	TDEPrincipalKey *curr_principal_key = NULL;
+	TDEPrincipalKey *new_principal_key = NULL;
 	LWLock	   *lock_files = tde_lwlock_enc_keys();
-	bool		is_dup_key = false;
+	bool		already_has_key = false;
+	GenericKeyring *new_keyring;
+	const KeyInfo *keyInfo = NULL;
+	bool		success = true;
+
+	if (AllowInheritGlobalProviders == false && providerOid != dbOid)
+	{
+		ereport(ERROR,
+				(errmsg("Usage of global key providers is disabled. Enable it with pg_tde.inherit_global_providers = ON")));
+	}
 
 	/*
 	 * Try to get principal key from cache.
 	 */
 	LWLockAcquire(lock_files, LW_EXCLUSIVE);
 
-	principalKey = get_principal_key_from_cache(dbOid);
-	is_dup_key = (principalKey != NULL);
+	curr_principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+	already_has_key = (curr_principal_key != NULL);
 
-	/* TODO: Add the key in the cache? */
-	if (!is_dup_key)
-		is_dup_key = (pg_tde_get_principal_key_info(dbOid) != NULL);
-
-	if (!is_dup_key)
+	if (provider_name == NULL && !already_has_key)
 	{
-		const keyInfo *keyInfo = NULL;
-
-		principalKey = palloc(sizeof(TDEPrincipalKey));
-		principalKey->keyInfo.databaseId = dbOid;
-		principalKey->keyInfo.keyId.version = DEFAULT_PRINCIPAL_KEY_VERSION;
-		principalKey->keyInfo.keyringId = keyring->key_id;
-		strncpy(principalKey->keyInfo.keyId.name, key_name, TDE_KEY_NAME_LEN);
-		gettimeofday(&principalKey->keyInfo.creationTime, NULL);
-
-		keyInfo = load_latest_versioned_key_name(&principalKey->keyInfo, keyring, ensure_new_key);
-
-		if (keyInfo == NULL)
-			keyInfo = KeyringGenerateNewKeyAndStore(keyring, principalKey->keyInfo.keyId.versioned_name, INTERNAL_KEY_LEN, true);
-
-		if (keyInfo == NULL)
-		{
-			LWLockRelease(lock_files);
-
-			ereport(ERROR,
-					(errmsg("failed to retrieve principal key. Create one using pg_tde_set_principal_key before using encrypted tables.")));
-		}
-
-		principalKey->keyLength = keyInfo->data.len;
-
-		memcpy(principalKey->keyData, keyInfo->data.data, keyInfo->data.len);
-
-		save_principal_key_info(&principalKey->keyInfo);
-
-		/* XLog the new key */
-		XLogBeginInsert();
-		XLogRegisterData((char *) &principalKey->keyInfo, sizeof(TDEPrincipalKeyInfo));
-		XLogInsert(RM_TDERMGR_ID, XLOG_TDE_ADD_PRINCIPAL_KEY);
-
-		push_principal_key_to_cache(principalKey);
-	}
-
-	LWLockRelease(lock_files);
-
-	if (is_dup_key)
-	{
-		/*
-		 * Seems like just before we got the lock, the key was installed by
-		 * some other caller Throw an error and mover no
-		 */
+		LWLockRelease(lock_files);
 
 		ereport(ERROR,
-				(errcode(ERRCODE_DUPLICATE_OBJECT),
-				 errmsg("Principal key already exists for the database"),
-				 errhint("Use rotate_key interface to change the principal key")));
+				(errmsg("provider_name is a required parameter when creating the first principal key for a database")));
 	}
 
-	return principalKey;
-}
-
-/*
- * alter_keyprovider_for_principal_key:
- */
-TDEPrincipalKey *
-alter_keyprovider_for_principal_key(GenericKeyring *newKeyring, Oid dbOid)
-{
-    TDEPrincipalKeyInfo *principalKeyInfo = NULL;
-	TDEPrincipalKey *principal_key = NULL;
-
-	LWLock *lock_files = tde_lwlock_enc_keys();
-
-	Assert(newKeyring != NULL);
-	LWLockAcquire(lock_files, LW_EXCLUSIVE);
-
-	principalKeyInfo = pg_tde_get_principal_key_info(dbOid);
-
-	if (principalKeyInfo == NULL)
+	if (provider_name != NULL)
 	{
-        LWLockRelease(lock_files);
-        ereport(ERROR,
-			(errmsg("Principal key not set for the database"),
-				errhint("Use set_principal_key interface to set the principal key")));
-	}
-
-	if (newKeyring->key_id == principalKeyInfo->keyringId)
-	{
-        LWLockRelease(lock_files);
-        ereport(ERROR,
-                (errmsg("New key provider is same as the current key provider")));
-    }
-    /* update the key provider in principal key info */
-
-	ereport(DEBUG2,
-			(errmsg("Changing keyprovider ID from :%d to %d", principalKeyInfo->keyringId, newKeyring->key_id)));
-
-	principalKeyInfo->keyringId = newKeyring->key_id;
-
-	update_principal_key_info(principalKeyInfo);
-
-	/* XLog the new key*/
-	XLogBeginInsert();
-	XLogRegisterData((char *)principalKeyInfo, sizeof(TDEPrincipalKeyInfo));
-	XLogInsert(RM_TDERMGR_ID, XLOG_TDE_UPDATE_PRINCIPAL_KEY);
-
-	/* clear the cache as well */
-	clear_principal_key_cache(dbOid);
-
-	principal_key = GetPrincipalKey(dbOid, LW_EXCLUSIVE);
-
-	LWLockRelease(lock_files);
-
-	return principal_key;
-}
-
-bool
-SetPrincipalKey(const char *key_name, const char *provider_name, bool ensure_new_key)
-{
-	TDEPrincipalKey *principal_key = set_principal_key_with_keyring(key_name,
-																	GetKeyProviderByName(provider_name, MyDatabaseId),
-																	MyDatabaseId,
-																	ensure_new_key);
-
-	return (principal_key != NULL);
-}
-
-bool
-AlterPrincipalKeyKeyring(const char *provider_name)
-{
-    TDEPrincipalKey *principal_key = alter_keyprovider_for_principal_key(GetKeyProviderByName(provider_name, MyDatabaseId),
-                                                                    MyDatabaseId);
-
-    return (principal_key != NULL);
-}
-
-bool
-RotatePrincipalKey(TDEPrincipalKey *current_key, const char *new_key_name, const char *new_provider_name, bool ensure_new_key)
-{
-	TDEPrincipalKey new_principal_key;
-	const keyInfo *keyInfo = NULL;
-	GenericKeyring *keyring;
-	bool is_rotated;
-	MemoryContext keyRotateCtx;
-	MemoryContext oldCtx;
-
-	Assert(current_key != NULL);
-
-	keyRotateCtx = AllocSetContextCreate(CurrentMemoryContext,
-										 "TDE key rotation temporary context",
-										 ALLOCSET_DEFAULT_SIZES);
-	oldCtx = MemoryContextSwitchTo(keyRotateCtx);
-
-	/*
-	 * Let's set everything the same as the older principal key and update
-	 * only the required attributes.
-	 */
-	memcpy(&new_principal_key, current_key, sizeof(TDEPrincipalKey));
-
-	if (new_key_name == NULL)
-	{
-		new_principal_key.keyInfo.keyId.version++;
+		new_keyring = GetKeyProviderByName(provider_name, providerOid);
 	}
 	else
 	{
-		strncpy(new_principal_key.keyInfo.keyId.name, new_key_name, sizeof(new_principal_key.keyInfo.keyId.name));
-		new_principal_key.keyInfo.keyId.version = DEFAULT_PRINCIPAL_KEY_VERSION;
+		new_keyring = GetKeyProviderByID(curr_principal_key->keyInfo.keyringId,
+										 curr_principal_key->keyInfo.databaseId);
+	}
 
-		if (new_provider_name != NULL)
+	{
+		KeyringReturnCodes kr_ret;
+
+		keyInfo = KeyringGetKey(new_keyring, key_name, false, &kr_ret);
+
+		if (kr_ret != KEYRING_CODE_SUCCESS && kr_ret != KEYRING_CODE_RESOURCE_NOT_AVAILABLE)
 		{
-			new_principal_key.keyInfo.keyringId = GetKeyProviderByName(new_provider_name,
-																	   new_principal_key.keyInfo.databaseId)->key_id;
+			ereport(ERROR,
+					(errmsg("failed to retrieve principal key from keyring provider :\"%s\"", new_keyring->provider_name),
+					 errdetail("Error code: %d", kr_ret)));
+			return false;
 		}
 	}
 
-	/* We need a valid keyring structure */
-	keyring = GetKeyProviderByID(new_principal_key.keyInfo.keyringId,
-								 new_principal_key.keyInfo.databaseId);
-
-	keyInfo = load_latest_versioned_key_name(&new_principal_key.keyInfo, keyring, ensure_new_key);
-
-	if (keyInfo == NULL)
-		keyInfo = KeyringGenerateNewKeyAndStore(keyring, new_principal_key.keyInfo.keyId.versioned_name, INTERNAL_KEY_LEN, true);
-
-	if (keyInfo == NULL)
+	if (keyInfo != NULL && ensure_new_key)
 	{
+		LWLockRelease(lock_files);
+
+		pfree(new_keyring);
+
 		ereport(ERROR,
-				(errmsg("Failed to generate new key name")));
+				(errmsg("failed to create principal key: already exists")));
+
+		return false;
 	}
 
-	new_principal_key.keyLength = keyInfo->data.len;
+	if (strlen(key_name) >= sizeof(keyInfo->name))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("too long principal key name, maximum lenght is %ld bytes", sizeof(keyInfo->name) - 1)));
 
-	memcpy(new_principal_key.keyData, keyInfo->data.data, keyInfo->data.len);
-	is_rotated = pg_tde_perform_rotate_key(current_key, &new_principal_key);
-	if (is_rotated && !TDEisInGlobalSpace(current_key->keyInfo.databaseId))
+	if (keyInfo == NULL)
+		keyInfo = KeyringGenerateNewKeyAndStore(new_keyring, key_name, INTERNAL_KEY_LEN, true);
+
+	if (keyInfo == NULL)
 	{
-		clear_principal_key_cache(current_key->keyInfo.databaseId);
-		push_principal_key_to_cache(&new_principal_key);
+		LWLockRelease(lock_files);
+
+		pfree(new_keyring);
+
+		ereport(ERROR,
+				(errmsg("failed to retrieve/create principal key.")));
+
+		return false;
 	}
 
-	MemoryContextSwitchTo(oldCtx);
-	MemoryContextDelete(keyRotateCtx);
+	new_principal_key = palloc(sizeof(TDEPrincipalKey));
+	new_principal_key->keyInfo.databaseId = dbOid;
+	new_principal_key->keyInfo.keyringId = new_keyring->keyring_id;
+	memcpy(new_principal_key->keyInfo.name, keyInfo->name, TDE_KEY_NAME_LEN);
+	gettimeofday(&new_principal_key->keyInfo.creationTime, NULL);
+	new_principal_key->keyLength = keyInfo->data.len;
 
-	return is_rotated;
+	memcpy(new_principal_key->keyData, keyInfo->data.data, keyInfo->data.len);
+
+	if (!already_has_key)
+	{
+		/* First key created for the database */
+		create_principal_key_info(&new_principal_key->keyInfo);
+
+		/* XLog the new key */
+		XLogBeginInsert();
+		XLogRegisterData((char *) &new_principal_key->keyInfo, sizeof(TDEPrincipalKeyInfo));
+		XLogInsert(RM_TDERMGR_ID, XLOG_TDE_ADD_PRINCIPAL_KEY);
+
+		push_principal_key_to_cache(new_principal_key);
+	}
+	else
+	{
+		/* key rotation */
+		bool		is_rotated = pg_tde_perform_rotate_key(curr_principal_key, new_principal_key);
+
+		if (is_rotated && (!TDEisInGlobalSpace(curr_principal_key->keyInfo.databaseId)))
+		{
+			clear_principal_key_cache(curr_principal_key->keyInfo.databaseId);
+			push_principal_key_to_cache(new_principal_key);
+		}
+
+		success = is_rotated;
+	}
+
+	LWLockRelease(lock_files);
+
+	pfree(new_keyring);
+
+	return success;
 }
 
 /*
@@ -454,7 +401,7 @@ RotatePrincipalKey(TDEPrincipalKey *current_key, const char *new_key_name, const
 bool
 xl_tde_perform_rotate_key(XLogPrincipalKeyRotate *xlrec)
 {
-	bool ret;
+	bool		ret;
 
 	ret = pg_tde_write_map_keydata_files(xlrec->map_size, xlrec->buff, xlrec->keydata_size, &xlrec->buff[xlrec->map_size]);
 	clear_principal_key_cache(xlrec->databaseId);
@@ -462,82 +409,6 @@ xl_tde_perform_rotate_key(XLogPrincipalKeyRotate *xlrec)
 	return ret;
 }
 
-/*
-* Load the latest versioned key name for the principal key
-* If ensure_new_key is true, then we will keep on incrementing the version number
-* till we get a key name that is not present in the keyring
-*/
-keyInfo *
-load_latest_versioned_key_name(TDEPrincipalKeyInfo *principal_key_info, GenericKeyring *keyring, bool ensure_new_key)
-{
-	KeyringReturnCodes kr_ret;
-	keyInfo *keyInfo = NULL;
-	int	base_version = principal_key_info->keyId.version;
-
-	Assert(principal_key_info != NULL);
-	Assert(keyring != NULL);
-	Assert(strlen(principal_key_info->keyId.name) > 0);
-
-	/*
-	 * Start with the passed in version number We expect the name and the
-	 * version number are already properly initialized and contain the correct
-	 * values
-	 */
-	snprintf(principal_key_info->keyId.versioned_name, TDE_KEY_NAME_LEN,
-			 "%s_%d", principal_key_info->keyId.name, principal_key_info->keyId.version);
-
-	while (true)
-	{
-		keyInfo = KeyringGetKey(keyring, principal_key_info->keyId.versioned_name, false, &kr_ret);
-
-		/*
-		 * vault-v2 returns 404 (KEYRING_CODE_RESOURCE_NOT_AVAILABLE) when key
-		 * is not found
-		 */
-		if (kr_ret != KEYRING_CODE_SUCCESS && kr_ret != KEYRING_CODE_RESOURCE_NOT_AVAILABLE)
-		{
-			ereport(ERROR,
-					(errmsg("failed to retrieve principal key from keyring provider :\"%s\"", keyring->provider_name),
-					 errdetail("Error code: %d", kr_ret)));
-			return NULL;
-		}
-		if (keyInfo == NULL)
-		{
-			if (ensure_new_key == false)
-			{
-				/*
-				 * If ensure_key is false and we are not at the base version,
-				 * We should return the last existent version.
-				 */
-				if (base_version < principal_key_info->keyId.version)
-				{
-					/* Not optimal but keep the things simple */
-					principal_key_info->keyId.version -= 1;
-					snprintf(principal_key_info->keyId.versioned_name, TDE_KEY_NAME_LEN,
-							 "%s_%d", principal_key_info->keyId.name, principal_key_info->keyId.version);
-					keyInfo = KeyringGetKey(keyring, principal_key_info->keyId.versioned_name, false, &kr_ret);
-				}
-			}
-			return keyInfo;
-		}
-
-		principal_key_info->keyId.version++;
-		snprintf(principal_key_info->keyId.versioned_name, TDE_KEY_NAME_LEN, "%s_%d", principal_key_info->keyId.name, principal_key_info->keyId.version);
-
-		/*
-		 * Not really required. Just to break the infinite loop in case the
-		 * key provider is not behaving sane.
-		 */
-		if (principal_key_info->keyId.version > MAX_PRINCIPAL_KEY_VERSION_NUM)
-		{
-			ereport(ERROR,
-					(errmsg("failed to retrieve principal key. %d versions already exist", MAX_PRINCIPAL_KEY_VERSION_NUM)));
-			return NULL;
-		}
-	}
-	return NULL;				/* Just to keep compiler quite */
-}
-
 /*
  * Returns the provider ID of the keyring that holds the principal key
  * Return InvalidOid if the principal key is not set for the database
@@ -547,9 +418,9 @@ GetPrincipalKeyProviderId(void)
 {
 	TDEPrincipalKey *principalKey = NULL;
 	TDEPrincipalKeyInfo *principalKeyInfo = NULL;
-	Oid	keyringId = InvalidOid;
-	Oid	dbOid = MyDatabaseId;
-	LWLock *lock_files = tde_lwlock_enc_keys();
+	Oid			keyringId = InvalidOid;
+	Oid			dbOid = MyDatabaseId;
+	LWLock	   *lock_files = tde_lwlock_enc_keys();
 
 	LWLockAcquire(lock_files, LW_SHARED);
 
@@ -618,11 +489,12 @@ static void
 push_principal_key_to_cache(TDEPrincipalKey *principalKey)
 {
 	TDEPrincipalKey *cacheEntry = NULL;
-	Oid databaseId = principalKey->keyInfo.databaseId;
-	bool found = false;
+	Oid			databaseId = principalKey->keyInfo.databaseId;
+	bool		found = false;
 
 	cacheEntry = dshash_find_or_insert(get_principal_key_Hash(),
 									   &databaseId, &found);
+
 	if (!found)
 		memcpy(cacheEntry, principalKey, sizeof(TDEPrincipalKey));
 	dshash_release_lock(get_principal_key_Hash(), cacheEntry);
@@ -686,100 +558,152 @@ clear_principal_key_cache(Oid databaseId)
 /*
  * SQL interface to set principal key
  */
-PG_FUNCTION_INFO_V1(pg_tde_set_principal_key);
-Datum		pg_tde_set_principal_key(PG_FUNCTION_ARGS);
+
+Datum
+pg_tde_set_default_principal_key(PG_FUNCTION_ARGS)
+{
+	char	   *principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = PG_ARGISNULL(1) ? NULL : text_to_cstring(PG_GETARG_TEXT_PP(1));
+	bool		ensure_new_key = PG_GETARG_BOOL(2);
+
+	return pg_tde_set_principal_key_internal(principal_key_name, GS_DEFAULT, provider_name, ensure_new_key);
+}
 
 Datum
 pg_tde_set_principal_key(PG_FUNCTION_ARGS)
 {
-	char *principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
-	char *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
-	bool ensure_new_key = PG_GETARG_BOOL(2);
-	bool ret;
+	char	   *principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = PG_ARGISNULL(1) ? NULL : text_to_cstring(PG_GETARG_TEXT_PP(1));
+	bool		ensure_new_key = PG_GETARG_BOOL(2);
+
+	return pg_tde_set_principal_key_internal(principal_key_name, GS_LOCAL, provider_name, ensure_new_key);
+}
+
+Datum
+pg_tde_set_global_principal_key(PG_FUNCTION_ARGS)
+{
+	char	   *principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = PG_ARGISNULL(1) ? NULL : text_to_cstring(PG_GETARG_TEXT_PP(1));
+	bool		ensure_new_key = PG_GETARG_BOOL(2);
+
+	return pg_tde_set_principal_key_internal(principal_key_name, GS_GLOBAL, provider_name, ensure_new_key);
+}
+
+Datum
+pg_tde_set_server_principal_key(PG_FUNCTION_ARGS)
+{
+	char	   *principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *provider_name = PG_ARGISNULL(1) ? NULL : text_to_cstring(PG_GETARG_TEXT_PP(1));
+	bool		ensure_new_key = PG_GETARG_BOOL(2);
+
+	return pg_tde_set_principal_key_internal(principal_key_name, GS_SERVER, provider_name, ensure_new_key);
+}
+
+static Datum
+pg_tde_set_principal_key_internal(char *principal_key_name, enum global_status global, char *provider_name, bool ensure_new_key)
+{
+	bool		success;
+	Oid			providerOid = MyDatabaseId;
+	Oid			dbOid = MyDatabaseId;
+	TDEPrincipalKey *existingDefaultKey = NULL;
+	TDEPrincipalKey existingKeyCopy;
 
 	ereport(LOG, (errmsg("Setting principal key [%s : %s] for the database", principal_key_name, provider_name)));
-	ret = SetPrincipalKey(principal_key_name, provider_name, ensure_new_key);
-	PG_RETURN_BOOL(ret);
-}
 
-PG_FUNCTION_INFO_V1(pg_tde_alter_principal_key_keyring);
-Datum pg_tde_alter_principal_key_keyring(PG_FUNCTION_ARGS);
-
-Datum pg_tde_alter_principal_key_keyring(PG_FUNCTION_ARGS)
-{
-	char *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
-	bool ret;
-
-	ereport(LOG, (errmsg("Altering principal key provider to \"%s\" for the database", provider_name)));
-	ret = AlterPrincipalKeyKeyring(provider_name);
-	PG_RETURN_BOOL(ret);
-}
-
-/*
- * SQL interface for key rotation
- */
-PG_FUNCTION_INFO_V1(pg_tde_rotate_principal_key_internal);
-Datum
-pg_tde_rotate_principal_key_internal(PG_FUNCTION_ARGS)
-{
-	char *new_principal_key_name = NULL;
-	char *new_provider_name = NULL;
-	bool ensure_new_key;
-	bool is_global;
-	bool ret;
-	TDEPrincipalKey *current_key;
-	Oid	dbOid = MyDatabaseId;
-
-	if (!PG_ARGISNULL(0))
-		new_principal_key_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
-	if (!PG_ARGISNULL(1))
-		new_provider_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
-	ensure_new_key = PG_GETARG_BOOL(2);
-	is_global = PG_GETARG_BOOL(3);
-
-#ifdef PERCONA_EXT
-	if (is_global)
+	if (global == GS_GLOBAL)	/* using a global provider for the current
+								 * database */
 	{
-		dbOid = GLOBAL_DATA_TDE_OID;
+		providerOid = GLOBAL_DATA_TDE_OID;
 	}
-#endif
-
-	ereport(LOG, (errmsg("rotating principal key to [%s : %s] the for the %s",
-						 new_principal_key_name,
-						 new_provider_name,
-						 is_global ? "cluster" : "database")));
-
-	LWLockAcquire(tde_lwlock_enc_keys(), LW_EXCLUSIVE);
-	current_key = GetPrincipalKey(dbOid, LW_EXCLUSIVE);
-	ret = RotatePrincipalKey(current_key, new_principal_key_name, new_provider_name, ensure_new_key);
-	LWLockRelease(tde_lwlock_enc_keys());
-
-	PG_RETURN_BOOL(ret);
-}
-
-PG_FUNCTION_INFO_V1(pg_tde_principal_key_info_internal);
-Datum
-pg_tde_principal_key_info_internal(PG_FUNCTION_ARGS)
-{
-	Oid	dbOid = MyDatabaseId;
-	bool is_global = PG_GETARG_BOOL(0);
-
-	if (is_global)
+	if (global == GS_SERVER)	/* using a globla provider for the global
+								 * (wal) database */
 	{
+		providerOid = GLOBAL_DATA_TDE_OID;
 		dbOid = GLOBAL_DATA_TDE_OID;
 	}
 
-	return pg_tde_get_key_info(fcinfo, dbOid);
+	if (global == GS_DEFAULT)	/* using a globla provider for the default
+								 * encryption setting */
+	{
+		providerOid = GLOBAL_DATA_TDE_OID;
+		dbOid = DEFAULT_DATA_TDE_OID;
+
+		/* Do we already have a default key? If yes, look up the name of it */
+		LWLockAcquire(tde_lwlock_enc_keys(), LW_SHARED);
+		existingDefaultKey = GetPrincipalKeyNoDefault(dbOid, LW_SHARED);
+		if (existingDefaultKey != NULL)
+		{
+			memcpy(&existingKeyCopy, existingDefaultKey, sizeof(TDEPrincipalKey));
+		}
+		LWLockRelease(tde_lwlock_enc_keys());
+	}
+
+	success = set_principal_key_with_keyring(principal_key_name,
+											 provider_name,
+											 providerOid,
+											 dbOid,
+											 ensure_new_key);
+
+	if (global == GS_DEFAULT && existingDefaultKey != NULL)
+	{
+		/*
+		 * In the previous step, we marked a new default provider Now we have
+		 * to start using the new principal key in all databases
+		 *
+		 * This violates the "only make changes to the actual database"
+		 * principle,
+		 *
+		 * But there's no really good way to do this unless we want to move
+		 * this to a command line tool.
+		 */
+
+		TDEPrincipalKey *newDefaultKey = NULL;
+
+		LWLockAcquire(tde_lwlock_enc_keys(), LW_EXCLUSIVE);
+		newDefaultKey = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+		pg_tde_update_global_principal_key_everywhere(&existingKeyCopy, newDefaultKey);
+
+		LWLockRelease(tde_lwlock_enc_keys());
+	}
+
+	PG_RETURN_BOOL(success);
+}
+
+PG_FUNCTION_INFO_V1(pg_tde_principal_key_info);
+Datum
+pg_tde_principal_key_info(PG_FUNCTION_ARGS)
+{
+	return pg_tde_get_key_info(fcinfo, MyDatabaseId);
+}
+
+PG_FUNCTION_INFO_V1(pg_tde_global_principal_key_info);
+Datum
+pg_tde_global_principal_key_info(PG_FUNCTION_ARGS)
+{
+	return pg_tde_get_key_info(fcinfo, GLOBAL_DATA_TDE_OID);
+}
+
+Datum
+pg_tde_verify_principal_key(PG_FUNCTION_ARGS)
+{
+	return pg_tde_verify_principal_key_internal(MyDatabaseId);
+}
+
+Datum
+pg_tde_verify_global_principal_key(PG_FUNCTION_ARGS)
+{
+	return pg_tde_verify_principal_key_internal(GLOBAL_DATA_TDE_OID);
 }
 
 static Datum
 pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
 {
-	TupleDesc tupdesc;
-	Datum values[6];
-	bool isnull[6];
-	HeapTuple tuple;
-	Datum result;
+	TupleDesc	tupdesc;
+	Datum		values[6];
+	bool		isnull[6];
+	HeapTuple	tuple;
+	Datum		result;
 	TDEPrincipalKey *principal_key;
 	TimestampTz ts;
 	GenericKeyring *keyring;
@@ -791,7 +715,7 @@ pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
 				 errmsg("function returning record called in context that cannot accept type record")));
 
 	LWLockAcquire(tde_lwlock_enc_keys(), LW_SHARED);
-	principal_key = GetPrincipalKey(dbOid, LW_SHARED);
+	principal_key = GetPrincipalKeyNoDefault(dbOid, LW_SHARED);
 	LWLockRelease(tde_lwlock_enc_keys());
 	if (principal_key == NULL)
 	{
@@ -801,12 +725,12 @@ pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
 		PG_RETURN_NULL();
 	}
 
-	keyring = GetKeyProviderByID(principal_key->keyInfo.keyringId, dbOid);
+	keyring = GetKeyProviderByID(principal_key->keyInfo.keyringId, principal_key->keyInfo.databaseId);
 
 	/* Initialize the values and null flags */
 
 	/* TEXT: Principal key name */
-	values[0] = CStringGetTextDatum(principal_key->keyInfo.keyId.name);
+	values[0] = CStringGetTextDatum(principal_key->keyInfo.name);
 	isnull[0] = false;
 	/* TEXT: Keyring provider name */
 	if (keyring)
@@ -821,17 +745,11 @@ pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
 	values[2] = Int32GetDatum(principal_key->keyInfo.keyringId);
 	isnull[2] = false;
 
-	/* TEXT: Principal key versioned name */
-	values[3] = CStringGetTextDatum(principal_key->keyInfo.keyId.versioned_name);
-	isnull[3] = false;
-	/* INTEGERT: Principal key version */
-	values[4] = Int32GetDatum(principal_key->keyInfo.keyId.version);
-	isnull[4] = false;
 	/* TIMESTAMP TZ: Principal key creation time */
 	ts = (TimestampTz) principal_key->keyInfo.creationTime.tv_sec - ((POSTGRES_EPOCH_JDATE - UNIX_EPOCH_JDATE) * SECS_PER_DAY);
 	ts = (ts * USECS_PER_SEC) + principal_key->keyInfo.creationTime.tv_usec;
-	values[5] = TimestampTzGetDatum(ts);
-	isnull[5] = false;
+	values[3] = TimestampTzGetDatum(ts);
+	isnull[3] = false;
 
 	/* Form the tuple */
 	tuple = heap_form_tuple(tupdesc, values, isnull);
@@ -839,6 +757,10 @@ pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
 	/* Make the tuple into a datum */
 	result = HeapTupleGetDatum(tuple);
 
+#ifndef FRONTEND
+	pfree(keyring);
+#endif
+
 	PG_RETURN_DATUM(result);
 }
 #endif							/* FRONTEND */
@@ -847,16 +769,16 @@ pg_tde_get_key_info(PG_FUNCTION_ARGS, Oid dbOid)
  * Gets principal key form the keyring and pops it into cache if key exists
  * Caller should hold an exclusive tde_lwlock_enc_keys lock
  */
-static TDEPrincipalKey *
-get_principal_key_from_keyring(Oid dbOid)
+TDEPrincipalKey *
+get_principal_key_from_keyring(Oid dbOid, bool pushToCache)
 {
 	GenericKeyring *keyring;
 	TDEPrincipalKey *principalKey = NULL;
 	TDEPrincipalKeyInfo *principalKeyInfo = NULL;
-	const keyInfo *keyInfo = NULL;
+	const KeyInfo *keyInfo = NULL;
 	KeyringReturnCodes keyring_ret;
 
-	Assert(LWLockHeldByMeInMode(tde_lwlock_enc_keys(), LW_EXCLUSIVE));
+	/* Assert(LWLockHeldByMeInMode(tde_lwlock_enc_keys(), LW_EXCLUSIVE)); */
 
 	principalKeyInfo = pg_tde_get_principal_key_info(dbOid);
 	if (principalKeyInfo == NULL)
@@ -870,7 +792,7 @@ get_principal_key_from_keyring(Oid dbOid)
 		return NULL;
 	}
 
-	keyInfo = KeyringGetKey(keyring, principalKeyInfo->keyId.versioned_name, false, &keyring_ret);
+	keyInfo = KeyringGetKey(keyring, principalKeyInfo->name, false, &keyring_ret);
 
 	if (keyInfo == NULL)
 	{
@@ -886,17 +808,19 @@ get_principal_key_from_keyring(Oid dbOid)
 	Assert(dbOid == principalKey->keyInfo.databaseId);
 
 #ifndef FRONTEND
-    /* We don't store global space key in cache */
-    if (!TDEisInGlobalSpace(dbOid))
-    {
-        push_principal_key_to_cache(principalKey);
+	/* We don't store global space key in cache */
+	if (pushToCache && !(TDEisInGlobalSpace(dbOid)))
+	{
+		push_principal_key_to_cache(principalKey);
 
-        /* If we do store key in cache we want to return a cache reference 
-         * rather then a palloc'ed copy.
-         */
-        pfree(principalKey);
-        principalKey = get_principal_key_from_cache(dbOid);
-    }
+		/*
+		 * If we do store key in cache we want to return a cache reference
+		 * rather then a palloc'ed copy.
+		 */
+		pfree(principalKey);
+		principalKey = get_principal_key_from_cache(dbOid);
+	}
+	pfree(keyring);
 #endif
 
 	if (principalKeyInfo)
@@ -919,13 +843,17 @@ get_principal_key_from_keyring(Oid dbOid)
  * But if there is no key in the cache, we have to switch the lock
  * (LWLockRelease + LWLockAcquire) to LW_EXCLUSIVE mode to write the key to the
  * cache.
+ *
+ * TODO: this and the above function is a terrible interface, refactoring needed
+ * * we are most likely leaking memory in frontend code
+ * * if this returns an object from the cache, caller has to hold the lock until it uses the object
+ * * mode of lock is unclear after returning from the function
  */
 TDEPrincipalKey *
-GetPrincipalKey(Oid dbOid, LWLockMode lockMode)
+GetPrincipalKeyNoDefault(Oid dbOid, LWLockMode lockMode)
 {
-#ifndef FRONTEND
 	TDEPrincipalKey *principalKey = NULL;
-
+#ifndef FRONTEND
 	Assert(LWLockHeldByMeInMode(tde_lwlock_enc_keys(), lockMode));
 	/* We don't store global space key in cache */
 	if (!TDEisInGlobalSpace(dbOid))
@@ -945,5 +873,278 @@ GetPrincipalKey(Oid dbOid, LWLockMode lockMode)
 	}
 #endif
 
-	return get_principal_key_from_keyring(dbOid);
+	return get_principal_key_from_keyring(dbOid, true);
 }
+
+TDEPrincipalKey *
+GetPrincipalKey(Oid dbOid, LWLockMode lockMode)
+{
+	TDEPrincipalKey *principalKey = GetPrincipalKeyNoDefault(dbOid, lockMode);
+#ifndef FRONTEND
+	TDEPrincipalKey *newPrincipalKey = NULL;
+#endif
+
+	if (principalKey != NULL)
+	{
+		return principalKey;
+	}
+
+#ifndef FRONTEND
+
+	/* Lock is already updated to exclusive at this point */
+	principalKey = GetPrincipalKeyNoDefault(DEFAULT_DATA_TDE_OID, LW_EXCLUSIVE);
+
+	if (principalKey == NULL)
+	{
+		return NULL;
+	}
+
+	newPrincipalKey = palloc(sizeof(TDEPrincipalKey));
+	memcpy(newPrincipalKey, principalKey, sizeof(TDEPrincipalKey));
+	newPrincipalKey->keyInfo.databaseId = dbOid;
+
+	create_principal_key_info(&newPrincipalKey->keyInfo);
+
+	push_principal_key_to_cache(newPrincipalKey);
+
+	pfree(newPrincipalKey);
+
+	principalKey = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+#endif
+
+	return principalKey;
+}
+
+#ifndef FRONTEND
+
+static bool
+pg_tde_is_provider_used(Oid databaseOid, Oid providerId)
+{
+	bool		is_global = (databaseOid == GLOBAL_DATA_TDE_OID);
+	bool		is_default = (databaseOid == DEFAULT_DATA_TDE_OID);
+
+	LWLockAcquire(tde_lwlock_enc_keys(), LW_EXCLUSIVE);
+
+	if (is_global || is_default)
+	{
+		HeapTuple	tuple;
+		SysScanDesc scan;
+		Relation	rel;
+
+		/* First verify that the global/default oid doesn't use it */
+
+		Oid			dbOid = GLOBAL_DATA_TDE_OID;
+		TDEPrincipalKey *principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+		if (principal_key != NULL && providerId == principal_key->keyInfo.keyringId)
+		{
+			LWLockRelease(tde_lwlock_enc_keys());
+
+			return true;
+		}
+
+		dbOid = DEFAULT_DATA_TDE_OID;
+		principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+		if (principal_key != NULL && providerId == principal_key->keyInfo.keyringId)
+		{
+			LWLockRelease(tde_lwlock_enc_keys());
+
+			return true;
+		}
+
+		/* We have to verify that it isn't currently used by any database */
+
+
+		rel = table_open(DatabaseRelationId, AccessShareLock);
+
+		scan = systable_beginscan(rel, 0, false, NULL, 0, NULL);
+
+		while ((tuple = systable_getnext(scan)) != NULL)
+		{
+			if (!HeapTupleIsValid(tuple))
+			{
+				break;
+			}
+
+			dbOid = ((Form_pg_database) GETSTRUCT(tuple))->oid;
+
+			principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+			if (principal_key == NULL)
+			{
+				continue;
+			}
+
+			if (providerId == principal_key->keyInfo.keyringId && principal_key->keyInfo.databaseId == GLOBAL_DATA_TDE_OID)
+			{
+				systable_endscan(scan);
+				table_close(rel, AccessShareLock);
+				LWLockRelease(tde_lwlock_enc_keys());
+
+				return true;
+			}
+		}
+
+		systable_endscan(scan);
+		table_close(rel, AccessShareLock);
+		LWLockRelease(tde_lwlock_enc_keys());
+
+		return false;
+	}
+	else
+	{
+		/* database local provider, just verify that it isn't currently active */
+
+		TDEPrincipalKey *principal_key = GetPrincipalKeyNoDefault(databaseOid, LW_EXCLUSIVE);
+
+		LWLockRelease(tde_lwlock_enc_keys());
+
+		return principal_key != NULL && providerId == principal_key->keyInfo.keyringId;
+	}
+}
+
+static bool
+pg_tde_is_same_principal_key(TDEPrincipalKey *a, TDEPrincipalKey *b)
+{
+	return a != NULL && b != NULL && strncmp(a->keyInfo.name, b->keyInfo.name, PRINCIPAL_KEY_NAME_LEN) == 0 && a->keyInfo.keyringId == b->keyInfo.keyringId;
+}
+
+static void
+pg_tde_rotate_default_key_for_database(TDEPrincipalKey *oldKey, TDEPrincipalKey *newKeyTemplate)
+{
+	bool		is_rotated;
+
+	TDEPrincipalKey *newKey = palloc(sizeof(TDEPrincipalKey));
+
+	memcpy(newKey, newKeyTemplate, sizeof(TDEPrincipalKey));
+	newKey->keyInfo.databaseId = oldKey->keyInfo.databaseId;
+
+	/* key rotation */
+	is_rotated = pg_tde_perform_rotate_key(oldKey, newKey);
+
+	if (is_rotated && (!TDEisInGlobalSpace(newKey->keyInfo.databaseId)))
+	{
+		clear_principal_key_cache(oldKey->keyInfo.databaseId);
+		push_principal_key_to_cache(newKey);
+	}
+
+	pfree(newKey);
+}
+
+static void
+pg_tde_update_global_principal_key_everywhere(TDEPrincipalKey *oldKey, TDEPrincipalKey *newKey)
+{
+	HeapTuple	tuple;
+	SysScanDesc scan;
+	Relation	rel;
+	Oid			dbOid = GLOBAL_DATA_TDE_OID;
+	TDEPrincipalKey *principal_key;
+
+	/* First check the global oid */
+	principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+	if (pg_tde_is_same_principal_key(oldKey, principal_key))
+	{
+		pg_tde_rotate_default_key_for_database(principal_key, newKey);
+	}
+
+
+	/*
+	 * Take row exclusive lock, as we do not want anybody to create/drop a
+	 * database in parallel. If it happens, its not the end of the world, but
+	 * not ideal
+	 */
+	rel = table_open(DatabaseRelationId, RowExclusiveLock);
+
+	scan = systable_beginscan(rel, 0, false, NULL, 0, NULL);
+
+	while ((tuple = systable_getnext(scan)) != NULL)
+	{
+		dbOid = ((Form_pg_database) GETSTRUCT(tuple))->oid;
+
+		principal_key = GetPrincipalKeyNoDefault(dbOid, LW_EXCLUSIVE);
+
+		if (pg_tde_is_same_principal_key(oldKey, principal_key))
+		{
+			pg_tde_rotate_default_key_for_database(principal_key, newKey);
+		}
+
+	}
+
+	systable_endscan(scan);
+	table_close(rel, RowExclusiveLock);
+}
+
+Datum
+pg_tde_delete_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_delete_key_provider_internal(fcinfo, 0);
+}
+
+Datum
+pg_tde_delete_global_key_provider(PG_FUNCTION_ARGS)
+{
+	return pg_tde_delete_key_provider_internal(fcinfo, 1);
+}
+
+Datum
+pg_tde_delete_key_provider_internal(PG_FUNCTION_ARGS, int is_global)
+{
+	char	   *provider_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	Oid			db_oid = (is_global == 1) ? GLOBAL_DATA_TDE_OID : MyDatabaseId;
+	GenericKeyring *provider = GetKeyProviderByName(provider_name, db_oid);
+	int			provider_id;
+	bool		provider_used;
+
+	if (provider == NULL)
+	{
+		ereport(ERROR,
+				(errmsg("Keyring provider not found")));
+	}
+
+	provider_id = provider->keyring_id;
+	provider_used = pg_tde_is_provider_used(db_oid, provider_id);
+
+	pfree(provider);
+
+	if (provider_used)
+	{
+		ereport(ERROR,
+				(errmsg("Can't delete a provider which is currently in use")));
+	}
+
+	delete_key_provider_info(provider_id, db_oid, true);
+
+	PG_RETURN_VOID();
+}
+
+static bool
+pg_tde_verify_principal_key_internal(Oid databaseOid)
+{
+	TDEPrincipalKey *fromKeyring;
+	TDEPrincipalKey *fromCache;
+
+	LWLockAcquire(tde_lwlock_enc_keys(), LW_EXCLUSIVE);
+
+	fromKeyring = get_principal_key_from_keyring(databaseOid, false);
+	fromCache = get_principal_key_from_cache(databaseOid);
+
+	LWLockRelease(tde_lwlock_enc_keys());
+
+	if (fromKeyring == NULL)
+	{
+		ereport(ERROR,
+				(errmsg("Failed to retrieve key from keyring")));
+	}
+
+	if (fromCache != NULL && (fromKeyring->keyLength != fromCache->keyLength || memcmp(fromKeyring->keyData, fromCache->keyData, fromCache->keyLength) != 0))
+	{
+		ereport(ERROR,
+				(errmsg("Key returned by keyring and cached in pg_tde is different")));
+	}
+
+	PG_RETURN_VOID();
+}
+
+#endif
