@@ -14,15 +14,12 @@
 #include "postgres.h"
 #include "catalog/pg_tablespace_d.h"
 
-#include "access/pg_tde_tdemap.h"
-#include "catalog/tde_principal_key.h"
-
 /*
  * Needed for global data (WAL etc) keys identification in caches and storage.
  * We take Oids of the sql operators, so there is no overlap with the "real"
  * catalog objects possible.
  */
-#define GLOBAL_DATA_TDE_OID	607
+#define GLOBAL_DATA_TDE_OID	607 // TODO: why not repeat GLOBALTABLESPACE_OID ?
 #define XLOG_TDE_OID        608
 
 #define GLOBAL_SPACE_RLOCATOR(_obj_oid) (RelFileLocator) { \
@@ -31,8 +28,9 @@
 	_obj_oid \
 }
 
+/*  Needed for using the same default key for multiple databases */
+#define DEFAULT_DATA_TDE_OID	DEFAULTTABLESPACE_OID
+
 #define TDEisInGlobalSpace(dbOid) 	(dbOid == GLOBAL_DATA_TDE_OID)
 
-extern void TDEInitGlobalKeys(const char *dir);
-
-#endif /* TDE_GLOBAL_CATALOG_H */
+#endif							/* TDE_GLOBAL_CATALOG_H */
