@@ -13,12 +13,10 @@
 #ifndef FRONTEND
 #include "nodes/pg_list.h"
 
-extern Oid	get_tde_basic_table_am_oid(void);
-extern Oid	get_tde_table_am_oid(void);
 extern List *get_all_tde_tables(void);
 extern int	get_tde_tables_count(void);
-#endif /* !FRONTEND */
+#endif							/* !FRONTEND */
 
 extern void pg_tde_set_data_dir(const char *dir);
-extern char* pg_tde_get_tde_data_dir(void);
-#endif /* PG_TDE_UTILS_H */
+extern char *pg_tde_get_tde_data_dir(void);
+#endif							/* PG_TDE_UTILS_H */
