@@ -9,25 +9,25 @@
 
 #include "postgres.h"
 #include "nodes/parsenodes.h"
-
-typedef enum TdeCreateEventType
-{
-	TDE_UNKNOWN_CREATE_EVENT,
-	TDE_TABLE_CREATE_EVENT,
-	TDE_INDEX_CREATE_EVENT
-} TdeCreateEventType;
+#include "access/transam.h"
 
 typedef struct TdeCreateEvent
 {
-	TdeCreateEventType eventType;	/* DDL statement type */
-	bool encryptMode;	/* true when the table uses encryption */
-	Oid baseTableOid;	/* Oid of table on which index is being
+	FullTransactionId tid;		/* transaction id of the last event trigger,
+								 * or 0 */
+	bool		encryptMode;	/* true when the table uses encryption */
+	Oid			baseTableOid;	/* Oid of table on which index is being
 								 * created on. For create table statement this
 								 * contains InvalidOid */
-	RangeVar *relation;		/* Reference to the parsed relation from
+	RangeVar   *relation;		/* Reference to the parsed relation from
 								 * create statement */
+	bool		alterAccessMethodMode;	/* during ALTER ... SET ACCESS METHOD,
+										 * new file permissions shouldn't be
+										 * based on earlier encryption status. */
 } TdeCreateEvent;
 
 extern TdeCreateEvent *GetCurrentTdeCreateEvent(void);
 
+extern void validateCurrentEventTriggerState(bool mightStartTransaction);
+
 #endif
