@@ -84,12 +84,13 @@ typedef struct _MdfdVec
 } MdfdVec;
 
 static MemoryContext MdCxt;		/* context for all MdfdVec objects */
-SMgrId MdSMgrId;
+SMgrId		MdSMgrId;
 
 typedef struct MdSMgrRelationData
 {
 	/* parent data */
 	SMgrRelationData reln;
+
 	/*
 	 * for md.c; per-fork arrays of the number of open segments
 	 * (md_num_open_segs) and the segments themselves (md_seg_fds).
@@ -133,10 +134,11 @@ typedef MdSMgrRelationData *MdSMgrRelation;
 #define EXTENSION_DONT_OPEN			(1 << 5)
 
 
-void mdsmgr_register(void)
+void
+mdsmgr_register(void)
 {
 	/* magnetic disk */
-	f_smgr md_smgr = (f_smgr) {
+	f_smgr		md_smgr = (f_smgr) {
 		.name = MdSMgrName,
 		.smgr_init = mdinit,
 		.smgr_shutdown = NULL,
@@ -237,7 +239,8 @@ mdcreate(RelFileLocator relold, SMgrRelation reln, ForkNumber forknum, bool isRe
 	char	   *path;
 	File		fd;
 	MdSMgrRelation mdreln = (MdSMgrRelation) reln;
-	// Assert(reln->smgr_which == MdSMgrId);
+
+	/* Assert(reln->smgr_which == MdSMgrId); */
 
 	if (isRedo && mdreln->md_num_open_segs[forknum] > 0)
 		return;					/* created and opened already... */
@@ -728,6 +731,7 @@ void
 mdopen(SMgrRelation reln)
 {
 	MdSMgrRelation mdreln = (MdSMgrRelation) reln;
+
 	/* mark it not open */
 	for (int forknum = 0; forknum <= MAX_FORKNUM; forknum++)
 		mdreln->md_num_open_segs[forknum] = 0;
@@ -1086,6 +1090,7 @@ mdwriteback(SMgrRelation reln, ForkNumber forknum,
 			BlockNumber blocknum, BlockNumber nblocks)
 {
 	MdSMgrRelation mdreln = (MdSMgrRelation) reln;
+
 	Assert((io_direct_flags & IO_DIRECT_DATA) == 0);
 
 	/*
@@ -1198,20 +1203,22 @@ mdnblocks(SMgrRelation reln, ForkNumber forknum)
 
 /*
  * mdtruncate() -- Truncate relation to specified number of blocks.
+ *
+ * Guaranteed not to allocate memory, so it can be used in a critical section.
+ * Caller must have called smgrnblocks() to obtain curnblk while holding a
+ * sufficient lock to prevent a change in relation size, and not used any smgr
+ * functions for this relation or handled interrupts in between.  This makes
+ * sure we have opened all active segments, so that truncate loop will get
+ * them all!
  */
 void
-mdtruncate(SMgrRelation reln, ForkNumber forknum, BlockNumber nblocks)
+mdtruncate(SMgrRelation reln, ForkNumber forknum,
+		   BlockNumber curnblk, BlockNumber nblocks)
 {
-	BlockNumber curnblk;
 	BlockNumber priorblocks;
 	int			curopensegs;
 	MdSMgrRelation mdreln = (MdSMgrRelation) reln;
 
-	/*
-	 * NOTE: mdnblocks makes sure we have opened all active segments, so that
-	 * truncation loop will get them all!
-	 */
-	curnblk = mdnblocks(reln, forknum);
 	if (nblocks > curnblk)
 	{
 		/* Bogus request ... but no complaint if InRecovery */
@@ -1552,7 +1559,7 @@ _fdvec_resize(MdSMgrRelation reln,
 		reln->md_seg_fds[forknum] =
 			MemoryContextAlloc(MdCxt, sizeof(MdfdVec) * nseg);
 	}
-	else
+	else if (nseg > reln->md_num_open_segs[forknum])
 	{
 		/*
 		 * It doesn't seem worthwhile complicating the code to amortize
@@ -1564,6 +1571,16 @@ _fdvec_resize(MdSMgrRelation reln,
 			repalloc(reln->md_seg_fds[forknum],
 					 sizeof(MdfdVec) * nseg);
 	}
+	else
+	{
+		/*
+		 * We don't reallocate a smaller array, because we want mdtruncate()
+		 * to be able to promise that it won't allocate memory, so that it is
+		 * allowed in a critical section.  This means that a bit of space in
+		 * the array is now wasted, until the next time we add a segment and
+		 * reallocate.
+		 */
+	}
 
 	reln->md_num_open_segs[forknum] = nseg;
 }
