@@ -66,12 +66,12 @@
 
 static f_smgr *smgrsw;
 
-static int NSmgr = 0;
+static int	NSmgr = 0;
 
 static Size LargestSMgrRelationSize = 0;
 
-char *storage_manager_string;
-SMgrId storage_manager_id;
+char	   *storage_manager_string;
+SMgrId		storage_manager_id;
 
 /*
  * Each backend has a hashtable that stores all extant SMgrRelation objects.
@@ -88,7 +88,7 @@ static void smgrdestroy(SMgrRelation reln);
 SMgrId
 smgr_register(const f_smgr *smgr, Size smgrrelation_size)
 {
-	SMgrId my_id;
+	SMgrId		my_id;
 	MemoryContext old;
 
 	if (process_shared_preload_libraries_done)
@@ -202,6 +202,7 @@ smgropen(RelFileLocator rlocator, ProcNumber backend)
 	{
 		/* First time through: initialize the hash table */
 		HASHCTL		ctl;
+
 		LargestSMgrRelationSize = MAXALIGN(LargestSMgrRelationSize);
 		Assert(NSmgr > 0);
 
@@ -691,14 +692,39 @@ smgrnblocks_cached(SMgrRelation reln, ForkNumber forknum)
  * smgrtruncate() -- Truncate the given forks of supplied relation to
  *					 each specified numbers of blocks
  *
+ * Backward-compatible version of smgrtruncate2() for the benefit of external
+ * callers.  This version isn't used in PostgreSQL core code, and can't be
+ * used in a critical section.
+ */
+void
+smgrtruncate(SMgrRelation reln, ForkNumber *forknum, int nforks,
+			 BlockNumber *nblocks)
+{
+	BlockNumber old_nblocks[MAX_FORKNUM + 1];
+
+	for (int i = 0; i < nforks; ++i)
+		old_nblocks[i] = smgrnblocks(reln, forknum[i]);
+
+	smgrtruncate2(reln, forknum, nforks, old_nblocks, nblocks);
+}
+
+/*
+ * smgrtruncate2() -- Truncate the given forks of supplied relation to
+ *					  each specified numbers of blocks
+ *
  * The truncation is done immediately, so this can't be rolled back.
  *
  * The caller must hold AccessExclusiveLock on the relation, to ensure that
  * other backends receive the smgr invalidation event that this function sends
- * before they access any forks of the relation again.
+ * before they access any forks of the relation again.  The current size of
+ * the forks should be provided in old_nblocks.  This function should normally
+ * be called in a critical section, but the current size must be checked
+ * outside the critical section, and no interrupts or smgr functions relating
+ * to this relation should be called in between.
  */
 void
-smgrtruncate(SMgrRelation reln, ForkNumber *forknum, int nforks, BlockNumber *nblocks)
+smgrtruncate2(SMgrRelation reln, ForkNumber *forknum, int nforks,
+			  BlockNumber *old_nblocks, BlockNumber *nblocks)
 {
 	int			i;
 
@@ -726,7 +752,8 @@ smgrtruncate(SMgrRelation reln, ForkNumber *forknum, int nforks, BlockNumber *nb
 		/* Make the cached size is invalid if we encounter an error. */
 		reln->smgr_cached_nblocks[forknum[i]] = InvalidBlockNumber;
 
-		smgrsw[reln->smgr_which].smgr_truncate(reln, forknum[i], nblocks[i]);
+		smgrsw[reln->smgr_which].smgr_truncate(reln, forknum[i],
+											   old_nblocks[i], nblocks[i]);
 
 		/*
 		 * We might as well update the local smgr_cached_nblocks values. The
