@@ -33,7 +33,9 @@
 #include "storage/bufpage.h"
 
 #ifdef PERCONA_EXT
-#include "access/pg_tde_xlog_encrypt_fe.h"
+#include "access/pg_tde_fe_init.h"
+#include "access/pg_tde_xlog_encrypt.h"
+#include "catalog/tde_global_space.h"
 #endif
 
 /*
@@ -248,7 +250,26 @@ search_directory(const char *directory, const char *fname)
 		PGAlignedXLogBlock buf;
 		int			r;
 
+#ifdef PERCONA_EXT
+		off_t		fsize;
+		TimeLineID	tli;
+		XLogSegNo	segno;
+
+		/*
+		 * WalSegSz extracted from the first page header but it might be
+		 * encrypted. But we need to know the segment seize to decrypt it
+		 * (it's required for encryption offset calculations). So we get the
+		 * segment size from the file's actual size.
+		 * XLogLongPageHeaderData->xlp_seg_size there is "just as a
+		 * cross-check" anyway.
+		 */
+		fsize = lseek(fd, 0, SEEK_END);
+		XLogFromFileName(fname, &tli, &segno, fsize);
+
+		r = xlog_smgr->seg_read(fd, buf.data, XLOG_BLCKSZ, 0, tli, segno, fsize);
+#else
 		r = read(fd, buf.data, XLOG_BLCKSZ);
+#endif
 		if (r == XLOG_BLCKSZ)
 		{
 			XLogLongPageHeader longhdr = (XLogLongPageHeader) buf.data;
@@ -765,34 +786,34 @@ usage(void)
 	printf(_("  %s [OPTION]... [STARTSEG [ENDSEG]]\n"), progname);
 	printf(_("\nOptions:\n"));
 	printf(_("  -b, --bkp-details         output detailed information about backup blocks\n"));
- 	printf(_("  -B, --block=N             with --relation, only show records that modify block N\n"));
- 	printf(_("  -e, --end=RECPTR          stop reading at WAL location RECPTR\n"));
- 	printf(_("  -f, --follow              keep retrying after reaching end of WAL\n"));
- 	printf(_("  -F, --fork=FORK           only show records that modify blocks in fork FORK;\n"
- 			 "                            valid names are main, fsm, vm, init\n"));
- 	printf(_("  -n, --limit=N             number of records to display\n"));
- 	printf(_("  -p, --path=PATH           directory in which to find WAL segment files or a\n"
- 			 "                            directory with a ./pg_wal that contains such files\n"
- 			 "                            (default: current directory, ./pg_wal, $PGDATA/pg_wal)\n"));
+	printf(_("  -B, --block=N             with --relation, only show records that modify block N\n"));
+	printf(_("  -e, --end=RECPTR          stop reading at WAL location RECPTR\n"));
+	printf(_("  -f, --follow              keep retrying after reaching end of WAL\n"));
+	printf(_("  -F, --fork=FORK           only show records that modify blocks in fork FORK;\n"
+			 "                            valid names are main, fsm, vm, init\n"));
+	printf(_("  -n, --limit=N             number of records to display\n"));
+	printf(_("  -p, --path=PATH           directory in which to find WAL segment files or a\n"
+			 "                            directory with a ./pg_wal that contains such files\n"
+			 "                            (default: current directory, ./pg_wal, $PGDATA/pg_wal)\n"));
 #ifdef PERCONA_EXT
- 	printf(_("  -k, --keyring-path=PATH   directory in which to find keyring config files for WAL\n"
- 			 "                            such files are pg_tde.map, pg_tde.dat, and pg_tde_keyrings\n"
- 			 "                            (it will not try to decrypt WAL if not set)\n"));
+	printf(_("  -k, --keyring-path=PATH   directory in which to find keyring config files for WAL\n"
+			 "                            such files are pg_tde.map, pg_tde.dat, and pg_tde_keyrings\n"
+			 "                            (it will not try to decrypt WAL if not set)\n"));
 #endif
- 	printf(_("  -q, --quiet               do not print any output, except for errors\n"));
- 	printf(_("  -r, --rmgr=RMGR           only show records generated by resource manager RMGR;\n"
- 			 "                            use --rmgr=list to list valid resource manager names\n"));
- 	printf(_("  -R, --relation=T/D/R      only show records that modify blocks in relation T/D/R\n"));
- 	printf(_("  -s, --start=RECPTR        start reading at WAL location RECPTR\n"));
- 	printf(_("  -t, --timeline=TLI        timeline from which to read WAL records\n"
- 			 "                            (default: 1 or the value used in STARTSEG)\n"));
- 	printf(_("  -V, --version             output version information, then exit\n"));
- 	printf(_("  -w, --fullpage            only show records with a full page write\n"));
- 	printf(_("  -x, --xid=XID             only show records with transaction ID XID\n"));
- 	printf(_("  -z, --stats[=record]      show statistics instead of records\n"
- 			 "                            (optionally, show per-record statistics)\n"));
- 	printf(_("  --save-fullpage=DIR       save full page images to DIR\n"));
- 	printf(_("  -?, --help                show this help, then exit\n"));
+	printf(_("  -q, --quiet               do not print any output, except for errors\n"));
+	printf(_("  -r, --rmgr=RMGR           only show records generated by resource manager RMGR;\n"
+			 "                            use --rmgr=list to list valid resource manager names\n"));
+	printf(_("  -R, --relation=T/D/R      only show records that modify blocks in relation T/D/R\n"));
+	printf(_("  -s, --start=RECPTR        start reading at WAL location RECPTR\n"));
+	printf(_("  -t, --timeline=TLI        timeline from which to read WAL records\n"
+			 "                            (default: 1 or the value used in STARTSEG)\n"));
+	printf(_("  -V, --version             output version information, then exit\n"));
+	printf(_("  -w, --fullpage            only show records with a full page write\n"));
+	printf(_("  -x, --xid=XID             only show records with transaction ID XID\n"));
+	printf(_("  -z, --stats[=record]      show statistics instead of records\n"
+			 "                            (optionally, show per-record statistics)\n"));
+	printf(_("  --save-fullpage=DIR       save full page images to DIR\n"));
+	printf(_("  -?, --help                show this help, then exit\n"));
 	printf(_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
 	printf(_("%s home page: <%s>\n"), PACKAGE_NAME, PACKAGE_URL);
 }
@@ -951,8 +972,8 @@ main(int argc, char **argv)
 				break;
 #ifdef PERCONA_EXT
 			case 'k':
- 				kringdir = pg_strdup(optarg);
- 				break;
+				kringdir = pg_strdup(optarg);
+				break;
 #endif
 			case 'q':
 				config.quiet = true;
@@ -1127,13 +1148,15 @@ main(int argc, char **argv)
 	}
 
 #ifdef PERCONA_EXT
- 	/* 
- 	 * Make possible to read ecrypted WAL
- 	 */
- 	if (kringdir != NULL)
- 	{
- 		TDE_XLOG_INIT(kringdir);
- 	}
+
+	/*
+	 * Make possible to read ecrypted WAL
+	 */
+	if (kringdir != NULL)
+	{
+		pg_tde_fe_init(kringdir);
+		TDEXLogSmgrInit();
+	}
 #endif
 
 	if (config.save_fullpage_path != NULL)
