@@ -14,15 +14,16 @@
 
 extern const PGDLLIMPORT int percona_api_version;
 
-static inline bool check_percona_api_version(void)
+static inline bool
+check_percona_api_version(void)
 {
-    if (PERCONA_API_VERSION != percona_api_version)
-    {
-        elog(FATAL, "Percona API version mismatch, the extension was built against a different PostgreSQL version!");
-        return false;
-    }
+	if (PERCONA_API_VERSION != percona_api_version)
+	{
+		elog(FATAL, "Percona API version mismatch, the extension was built against a different PostgreSQL version!");
+		return false;
+	}
 
-    return true;
+	return true;
 }
 
-#endif // PERCONA__H__
\ No newline at end of file
+#endif
